# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager, suppress
from typing import Any

from ..exceptions import (
    FailedExpectingEndOfLine,
    FailedExpectingEndOfText,
    FailedLookahead,
    FailedMeta,
    FailedParse,
    FailedPattern,
    FailedToken,
    OptionSucceeded,
    ParseException,
)
from ..util import boundcall, deprecated, left_assoc, regexpp, right_assoc
from .cst import closedlist, cstfinal
from .ctx import Func
from .ctxlib import (
    ChoiceContext,
    ExpContext,
    ExpWithSepContext,
    LoopContext,
    LoopWithSepContext,
)
from .engine import ParserEngine
from .state import _AT_


class ParseContext(ParserEngine):
    # bw compatibility
    @deprecated(replacement=ParserEngine.newexcept)
    def _error(
        self,
        item: Any,
        exclass: type[FailedParse] = FailedParse,
    ) -> FailedParse:
        raise self.newexcept(item, exclass)

    @property
    def last_node(self) -> Any:
        return self.state.last_node

    # bw compatibility
    @deprecated(replacement=ParserEngine.setname)
    def name_last_node(self, name: str):  # backwards compatibility
        self.setname(name)

    # bw compatibility
    @deprecated(replacement=ParserEngine.addname)
    def add_last_node_to_name(self, name: str):  # backwards compatibility
        self.addname(name)

    def fail(self):
        self.next_token()
        raise self.newexcept('fail')

    _fail = fail

    def alert(self, message: str, level: int) -> None:
        self.next_token()
        self.tracer.trace_match(self, f'{"^" * level}`{message}`', failed=True)
        # note: capture=False, nothing appended to state
        message = self.constant(message, capture=False)
        self.states.alert(level=level, message=message)

    _aler = alert

    def token(self, token: str) -> str:
        self.next_token()
        if self.cursor.match(token) is None:
            self.tracer.trace_match(self, token, failed=True)
            raise self.newexcept(token, excls=FailedToken)
        self.tracer.trace_match(self, token)
        self.state.append(token)
        return token

    _token = token

    def pattern(self, pattern: str) -> Any:
        if (token := self.cursor.matchre(pattern)) is None:
            self.tracer.trace_match(self, '', pattern, failed=True)
            raise self.newexcept(f'Expecting {regexpp(pattern)}', excls=FailedPattern)
        self.tracer.trace_match(self, token, pattern)
        self.state.append(token)
        return token

    _pattern = pattern

    def matchname(self) -> str | None:
        self.next_token()
        if (token := self.cursor.matchname()) is None:
            self.tracer.trace_match(self, '', '@name', failed=True)
            raise self.newexcept('Expecting @name', excls=FailedMeta)
        self.tracer.trace_match(self, token, '@name')
        self.state.append(token)
        return token

    _matchname = matchname

    def matchint(self) -> int | None:
        self.next_token()
        if (token := self.cursor.matchint()) is None:
            self.tracer.trace_match(self, '', '@int', failed=True)
            raise self.newexcept('Expecting @int', excls=FailedMeta)
        self.tracer.trace_match(self, str(token), '@int')
        self.state.append(token)
        return token

    _matchint = matchint

    def matchuint(self) -> int | None:
        self.next_token()
        if (token := self.cursor.matchuint()) is None:
            self.tracer.trace_match(self, '', '@int', failed=True)
            raise self.newexcept('Expecting @int', excls=FailedMeta)
        self.tracer.trace_match(self, str(token), '@int')
        self.state.append(token)
        return token

    _matchuint = matchuint

    def matchfloat(self) -> float | None:
        self.next_token()
        if (token := self.cursor.matchfloat()) is None:
            self.tracer.trace_match(self, '', '@float', failed=True)
            raise self.newexcept('Expecting @float', excls=FailedMeta)
        self.tracer.trace_match(self, str(token), '@float')
        self.state.append(token)
        return token

    _matchfloat = matchfloat

    def matchbool(self) -> bool | None:
        self.next_token()
        if (token := self.cursor.matchbool()) is None:
            self.tracer.trace_match(self, '', '@bool', failed=True)
            raise self.newexcept('Expecting @bool', excls=FailedMeta)
        self.tracer.trace_match(self, str(token), '@bool')
        self.state.append(token)
        return token

    _matchbool = matchbool

    def eof(self) -> bool:
        return self.cursor.atend()

    def eol(self) -> bool:
        return self.cursor.ateol()

    def eofcheck(self) -> None:
        self.next_token()
        if not self.cursor.atend():
            raise self.newexcept(
                'Expecting end of text',
                excls=FailedExpectingEndOfText,
            )

    _check_eof = eofcheck

    def eolcheck(self):
        if not self.state.cursor.matcheol():
            raise self.newexcept(
                'Expecting end of line',
                excls=FailedExpectingEndOfLine,
            )

    def _no_more_options(self) -> bool:
        # NOTE: Legacy. Used in previous versions of the parser generator
        return True

    @contextmanager
    def option(self) -> Any:
        self.states.push()
        try:
            yield
            self.states.merge()
            raise OptionSucceeded()
        except FailedParse:
            if self.states.undo().cutseen:
                raise

    _option = option

    @contextmanager
    def choice(self) -> Generator[ChoiceContext, Any, Any]:
        chc = ChoiceContext(self)
        with suppress(OptionSucceeded):
            yield chc
            chc.parse(self)

    _choice = choice

    @contextmanager
    def optional(self) -> Any:
        # NOTE
        #   For reimplementors.
        #   These few lines of code define the complete semantics of Cut (~).
        #   It is contained to an Option/Alt.
        #   The .push() provides a new state with cutseen==False.
        #   The .merge()/.undo() contain the Cut so an outer Option cannot use it.
        #   The OptionSucceeded exception is just the legacy way of the library
        #   to manage the control flow on success (some versions of the library
        #   have just returned).
        self.states.push()
        try:
            yield
            self.states.merge()
        except FailedParse:
            if self.states.undo().cutseen:
                raise

    _optional = optional

    @contextmanager
    def group(self) -> Any:
        yield

    _group = group

    def groupexp(self, exp: Func) -> Any:
        return exp(self)

    @contextmanager
    def skipgroup(self) -> Any:
        with self.statescope(merge=False):
            yield

    _skipgroup = skipgroup

    @contextmanager
    def if_(self) -> Any:
        self.states.push()
        try:
            yield
        finally:
            self.states.undo()

    _if = if_

    @contextmanager
    def ifnot_(self) -> Any:
        try:
            with self.if_():
                yield
        except ParseException:
            pass
        else:
            raise self.newexcept('', excls=FailedLookahead)

    _ifnot = ifnot_

    @contextmanager
    def nameset(self, name: str) -> Any:
        yield
        self.state.nameset(name)

    _setname = nameset

    @contextmanager
    def nameadd(self, name: str) -> Any:
        yield
        self.state.nameadd(name)

    _addname = nameadd

    @contextmanager
    def result(self) -> Any:
        yield
        self.state.nameset(_AT_)

    @contextmanager
    def resultadd(self) -> Any:
        yield
        self.state.nameadd(_AT_)

    def expcall(self, exp: Func) -> Any:
        try:
            return exp(self)
        except TypeError as e:
            if "arguments" in str(e):
                return boundcall(exp, {}, self)
            raise

    def isolate(self, exp: Func) -> Any:
        self.states.push()
        try:
            self.expcall(exp)
            return cstfinal(self.cst)
        finally:
            ast = self.ast
            self.states.pop()
            self.ast = ast

    _isolate = isolate

    def repeat(
        self,
        exp: Func,
        prefix: Func | None = None,
        omitsep: bool = False,
    ) -> None:
        while True:
            with suppress(OptionSucceeded):
                with self.option():
                    p = self.pos

                    if prefix:
                        pcst = self.isolate(prefix)
                        if not omitsep:
                            self.state.append(pcst)
                        self.cut()

                    cst = self.isolate(exp)
                    self.state.append(cst)

                    if self.pos == p:
                        raise self.newexcept(
                            f'{self.repeat.__name__} matched on no input',
                        )
                # note: dit not match sep? exp, so quit
                break

    def closure(
        self,
        exp: Func,
        sep: Func | None = None,
        omitsep: bool = False,
    ) -> Any:
        with self.statescope():
            self.cst = []
            with self.optional():
                self.expcall(exp)
                self.cst = [self.cst]
                self.repeat(exp, prefix=sep, omitsep=omitsep)
            self.cst = cst = closedlist(self.cst)
            return cst

    _closure = closure

    def positive_closure(
        self,
        exp: Func,
        sep: Func | None = None,
        omitsep: bool = False,
    ) -> Any:
        with self.statescope():
            self.expcall(exp)
            self.cst = [self.cst]
            self.repeat(exp, prefix=sep, omitsep=omitsep)
            self.cst = cst = closedlist(self.cst)
            return cst

    _positive_closure = positive_closure

    @contextmanager
    def loopopt(self) -> Any:
        cl = LoopContext(self)
        yield cl
        self.closure(cl.func, omitsep=False)

    @contextmanager
    def loopplus(self) -> Any:
        cl = LoopContext(self, plus=True)
        yield cl
        self.positive_closure(cl.func, omitsep=False)

    def empty(self) -> list:
        cst = closedlist([])
        self.state.append(cst)
        return cst

    def _empty_closure(self) -> list:
        return self.empty()

    @contextmanager
    def gatheropt(self) -> Any:
        cl = LoopWithSepContext(self, plus=False, omitsep=True)
        yield cl
        self.gather(cl.func, cl.sep_func)

    @contextmanager
    def gatherplus(self) -> Any:
        cl = LoopWithSepContext(self, plus=True, omitsep=True)
        yield cl
        self.positive_gather(cl.func, cl.sep_func)

    def gather(self, exp: Func, sep: Func) -> Any:
        return self.closure(exp, sep=sep, omitsep=True)

    _gather = gather

    def positive_gather(self, exp: Func, sep: Func) -> Any:
        return self.positive_closure(exp, sep=sep, omitsep=True)

    _positive_gather = positive_gather

    @contextmanager
    def joinopt(self) -> Any:
        cl = ExpWithSepContext(self)
        yield cl
        self.join(cl.func, cl.sep_func)

    @contextmanager
    def joinplus(self) -> Any:
        cl = ExpWithSepContext(self)
        yield cl
        self.positive_join(cl.func, cl.sep_func)

    def join(self, exp: Func, sep: Func) -> Any:
        return self.closure(exp, sep=sep, omitsep=False)

    _join = join

    def positive_join(self, exp: Func, sep: Func) -> Any:
        return self.positive_closure(exp, sep=sep, omitsep=False)

    _positive_join = positive_join

    @contextmanager
    def joinleft(self) -> Any:
        cl = ExpWithSepContext(self)
        yield cl
        self.left_join(cl.func, cl.sep_func)

    @contextmanager
    def joinright(self) -> Any:
        cl = ExpWithSepContext(self)
        yield cl
        self._right_join(cl.func, cl.sep_func)

    def left_join(self, exp: Func, sep: Func) -> Any:
        self.cst = left_assoc(self.positive_join(exp, sep))
        return self.cst

    _left_join = left_join

    def right_join(self, exp: Func, sep: Func) -> Any:
        self.cst = right_assoc(self.positive_join(exp, sep))
        return self.cst

    _right_join = right_join

    def void(self) -> Any:
        self.next_token()
        return ()

    _void = void

    def dot(self) -> Any:
        c = self._next()
        if c is None:
            self.tracer.trace_match(self, c, failed=True)
            raise self.newexcept(c, excls=FailedToken) from None
        self.tracer.trace_match(self, c)
        self.state.append(c)
        return c

    _dot = dot

    @contextmanager
    def skipto(self) -> Any:
        cl = ExpContext(self)
        yield cl
        self.skip_to(cl.func)

    _skipto = skipto

    def skip_to(self, exp: Func) -> Any:
        while not self.eof():
            try:
                with self.if_():
                    self.expcall(exp)
            except FailedParse:
                pass
            else:
                break
            pos = self.pos
            self.next_token()
            if self.pos == pos:
                self._next()
        return self.expcall(exp)

    _skip_to = skip_to



# idea: off
# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Any


class FastProxy[T]:
    def __init__(self, obj: T, **other: Any) -> None:
        object.__setattr__(self, '_obj', obj)
        for key, value in other.items():
            name = f'_{key}' if not key.startswith('_') else key
            object.__setattr__(self, name, value)

    def __getattr__(self, name: str) -> Any:
        value = getattr(self._obj, name)
        if callable(value):
            unbound = getattr(type(self._obj), name)
            if hasattr(unbound, "__get__"):
                # This handles both methods and properties correctly
                bound = unbound.__get__(self._obj)
                object.__setattr__(self, name, bound)
                return bound

        return value

    def __setattr__(self, name: str, value: Any) -> None:
        if name == '_obj' or name not in vars(self._obj):
            super().__setattr__(name, value)
        else:
            setattr(self._obj, name, value)

    def __dir__(self) -> list[str]:
        return sorted(set(super().__dir__()) | set(dir(self._obj)))

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._obj!r}))'


if __name__ == '__main__':

    class C:
        def __init__(self):
            self._x = 0

        @property
        def x(self):
            return self._x

        @x.setter
        def x(self, value):
            self._x = value

        @property
        def dx(self) -> int:
            return 2 * self._x

        def incr(self) -> int:
            self.x += 1
            return self.x

        def whoami(self) -> str:
            return f'I am {type(self).__name__} {self!r}'

    c = C()
    p = FastProxy(c)

    print(p.incr())
    assert c.x == 1, c.x
    assert p.x == 1, p.x

    print(c.incr())
    assert c.x == 2, c.x
    assert p.x == 2, p.x

    class Over[C](FastProxy[C]):
        def __init__(self, obj: C):
            super().__init__(obj)
            object.__setattr__(self, 'x', None)
            self.y = 66

        @property
        def x(self):
            return 42

        @x.setter
        def x(self, value):
            _ = value

        def incr(self) -> int:
            return self._obj.incr()

    o = Over(c)
    print(o.incr())
    print(o.dx)

    print(c.whoami())
    print(o.whoami())

    print(type(c))
    print(type(o))

    print(c.incr)
    print(p.incr)
    print(o.incr)

    print(C.x)
    print(p.x)
    print(Over.x)

    print(c.incr())
    assert o.x == 42, o.x
    assert p.x == 4, p.x
    assert c.x == 4, c.x

    o.y = 22

    assert not hasattr(c, 'y')
    assert not hasattr(p, 'y')
    assert hasattr(o, 'y')
    assert o._obj is c

    print(vars(o))



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations


"""
Define the non-AST semantics parging.

A CST is used to gather the values returned by expressions into the value
returned from a grammar rule invocation.

If a CST is to be returned from a grammar rule invocation, it is first converted
to a `tuple` (elseqwhere) so it doesn't play these games.
"""

from typing import Any


class closedlist(list[Any]):
    """
    A `list` that is not a list.

    It's used by closures and friends to keep their results as values that
    don't get merged with other lists.
    """

    __setitem__ = None  # type: ignore
    __delitem__ = None  # type: ignore
    __iadd__ = None  # type: ignore
    insert = None  # type: ignore

    def __hash__(self) -> int:  # type: ignore
        return hash(tuple(self))


def islist(o: Any) -> bool:
    return not isinstance(o, closedlist) and isinstance(o, list)


def cstfinal(cst: Any) -> Any:
    return closedlist(cst) if islist(cst) else cst


def cstadd(cst: Any, node: Any) -> Any:
    """
    This is how the values of subexpressions are added to the result of the
    enclosing context: a rule, group, closure, ...
    :param cst: the previous CST value
    :param node: the node to add
    :param aslist: for rules that require that the final CST is a list
    :return: the new CST value
    """
    if cst is None:
        return node
    if islist(cst):
        return [*cst, node]
    return [cst, node]


def cstaddlist(cst: Any, node: Any) -> Any:
    """
    This is how the values of subexpressions are added to the result of the
    enclosing context: a rule, group, closure, ...
    :param cst: the previous CST value
    :param node: the node to add
    :param aslist: for rules that require that the final CST is a list
    :return: the new CST value
    """
    if cst is None:
        return [node]
    if islist(cst):
        return [*cst, node]
    return [cst, node]


def cstmerge(cst: Any, other: Any) -> Any:
    """
    Used to merge the CST of a subexpression into the CST of the enclosing
    context.
    :param cst: the "outer" CST
    :param other: the CST to merge
    :return: the merged CST
    """
    if other is None:
        return cst
    if cst is None:
        return other  # avoid shared state of lists
    if islist(other) and islist(cst):
        return cst + other
    if islist(other):
        return [cst, *other]
    if islist(cst):
        return [*cst, other]
    return [cst, other]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from io import StringIO

from ..input import LineInfo
from ..util.strtools import slicetowidth
from ..ztyle import Color, Style


MEMENTO_DEFAULT_COLOR = Color.stderr()


class _ColorSet:
    def __init__(self, color: Color = MEMENTO_DEFAULT_COLOR):
        self.err = Style(bold=True, fg=1, color=color)
        self.loc = Style(fg=4, color=color)
        self.gut = Style(color=color).blue().bold()
        self.ar = Style(color=color).red().dim()
        self.nam = Style(color=color).white().bold()
        self.msg = Style(color=color).white().bold()
        self.num = Style(color=color).white().dim()


def memento(
    msg: str,
    text: str,
    info: LineInfo,
    rulestack: list[str],
    color: Color = MEMENTO_DEFAULT_COLOR,
) -> str:
    c = _ColorSet(color)
    line, col = info.line, info.col
    source = info.source or '<unknown>'

    out = StringIO()
    s = Style(color=color)

    lines = text.splitlines()
    errmsg = f'{c.err("error:")} {c.msg(msg)}'
    print(errmsg, file=out)
    loc = s(f'[{line + 1}:{col + 1}]').dim()
    print(
        f'{c.gut("  ->")} {c.nam(source)}{loc}',
        file=out,
    )
    gut = c.gut("│")
    print(f'   {gut}', file=out)

    max_line_digits = len(str(line + 1))
    start_line_idx = max(0, line - 4)

    for i in range(start_line_idx, min(line + 1, len(lines))):
        current_line_num = i + 1
        content = lines[i].expandtabs()
        snum = c.num(f'{current_line_num:>{max_line_digits}}')
        print(
            f' {snum} {gut} {content}',
            file=out,
        )

    padding = ' ' * max(0, col)
    print(
        f' {" ":{max_line_digits + 1}}{gut}'
        f' {padding}{c.err("⌃ ")} {c.err(slicetowidth(msg, 40))}',
        file=out,
    )

    print(file=out)
    for call in rulestack:
        print(f'{c.ar("→")} {s(call).dim()}', file=out)

    return out.getvalue()



#!/usr/bin/env python3
# copyright: ignore
#
# CAVEAT UTILITOR
#
#  This file was automatically generated by 竜TatSu vX.YY.Z
#
#     https://pypi.python.org/pypi/tatsu/
#
#  Any changes you make to it will be overwritten the next time
#  the file is generated.
#
# noqa # type: ignore # ruff: noqa
# fmt: off

from __future__ import annotations

from typing import Any

from tatsu import decorators as tatsu
from tatsu.config import ParserConfig
from tatsu.contexts import Ctx
from tatsu.parsing import Parser, generic_main
from tatsu.input.buffer import Buffer
from tatsu.input.textlines import TextLines

KEYWORDS: set[str] = set()


class LISPTokenizer(TextLines):
    def __init__(
        self,
        text, /,
        config: ParserConfig | None = None,
        **settings,
    ) -> None:
        config = ParserConfig.new(
            config=config,
            whitespace=None,
            nameguard=None,
            ignorecase=False,
            namechars='',
            parseinfo=False,
            comments=r'None',
            eol_comments=r'None',
            keywords=KEYWORDS,
            start='start',
        )
        assert isinstance(config, ParserConfig)
        config = config.override(**settings)

        super().__init__(text, config=config)


class LISPBuffer(Buffer):  # NOTE: backwards compatibility
    def __init__(
        self,
        text, /,
        config: ParserConfig | None = None,
        **settings,
    ) -> None:
        config = ParserConfig.new(
            config=config,
            whitespace=None,
            nameguard=None,
            ignorecase=False,
            namechars='',
            parseinfo=False,
            comments=r'None',
            eol_comments=r'None',
            keywords=KEYWORDS,
            start='start',
        )
        assert isinstance(config, ParserConfig)
        config = config.override(**settings)

        super().__init__(text, config=config)


class LISPParser(Parser):
    def __init__(self, /, config: ParserConfig | None = None, **settings):
        config = ParserConfig.new(config, **settings)
        rulessource = LISPRules()
        assert isinstance(config, ParserConfig)

        super().__init__(rulessource, config=config)


class LISPRules:
    def __init__(self, /, config: ParserConfig | None = None, **settings):
        config = ParserConfig.new(
            config=config,
            whitespace=None,
            nameguard=None,
            ignorecase=False,
            namechars='',
            parseinfo=False,
            comments=r'None',
            eol_comments=r'None',
            keywords=KEYWORDS,
            start='start',
        )
        assert isinstance(config, ParserConfig)
        config = config.override(**settings)
        self._config = config


    @tatsu.rule
    def start(self, ctx: Ctx) -> Any:
        self.lisp(ctx)

    @tatsu.rule
    def lisp(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            @α.option
            def 〇(ctx: Ctx) -> Any:
                self.sexp(ctx)
            @α.option
            def 〇(ctx: Ctx) -> Any:
                self.list_(ctx)
            @α.option
            def 〇(ctx: Ctx) -> Any:
                self.symbol(ctx)
            @α.option
            def 〇(ctx: Ctx) -> Any:
                self.number(ctx)

    @tatsu.rule('SExp')
    def sexp(self, ctx: Ctx) -> Any:
        ctx.define(['cdr', 'cons'], [])
        ctx.token('(')
        with ctx.nameset('cons'):
            self.lisp(ctx)
        ctx.token('.')
        ctx.cut()
        with ctx.nameset('cdr'):
            self.lisp(ctx)
        ctx.token(')')

    @tatsu.rule('List')
    def list_(self, ctx: Ctx) -> Any:
        ctx.token('(')
        with ctx.nameset('@'):
            with ctx.loopopt() as cl:
                cl.expecting('<lisp>')

                @cl.exp
                def 〇(ctx: Ctx) -> Any:
                    self.lisp(ctx)
        ctx.token(')')

    @tatsu.rule('Symbol')
    def symbol(self, ctx: Ctx) -> Any:
        ctx.pattern(r'[_\w][_\w\d]*')

    @tatsu.rule('Number')
    def number(self, ctx: Ctx) -> Any:
        ctx.pattern(r'\d+(?:\.\d+)')


def main(filename, **kwargs):
    if not filename or filename == '-':
        import sys

        text = sys.stdin.read()
    else:
        import pathlib

        text = pathlib.Path(filename).read_text()

    parser = LISPParser()
    return parser.parse(text, filename=filename, **kwargs)


if __name__ == '__main__':
    import json
    from tatsu.util import asjson

    ast = generic_main(main, LISPParser, name='LISP')
    data = asjson(ast)
    print(json.dumps(data, indent=2))



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Iterable
from enum import Enum, auto

from ..basic import Cut
from ..choice import Choice
from ..model import Box, Model, Rule
from ..syntax import Call, Sequence


__all__ = ['mark_left_recursion']


def _callable_rule_ids(exp: Model, rule_index: dict[str, int]) -> list[int]:
    if isinstance(exp, Call):
        target = rule_index.get(exp.name)
        return [target] if target is not None else []

    if isinstance(exp, Choice):
        result: list[int] = []
        for opt in exp.options:
            result.extend(_callable_rule_ids(opt, rule_index))
        return result

    if isinstance(exp, Sequence):
        result = []
        for item in exp.sequence:
            if isinstance(item, Cut):
                continue
            result.extend(_callable_rule_ids(item, rule_index))
            if not _is_nullable_safe(item):
                break
        return result

    if isinstance(exp, Box):
        return _callable_rule_ids(exp.exp, rule_index)

    return []


def _is_nullable_safe(exp: Model) -> bool:
    if isinstance(exp, Call):
        return False

    if isinstance(exp, Sequence):
        return all(_is_nullable_safe(item) for item in exp.sequence)

    if isinstance(exp, Choice):
        return any(_is_nullable_safe(opt) for opt in exp.options)

    return exp.is_nullable()


def mark_left_recursion(rules: Iterable[Rule]) -> list[Rule]:
    class State(Enum):
        FIRST = auto()
        VISITING = auto()
        VISITED = auto()

    rules = list(rules)
    rule_index = {rule.name: i for i, rule in enumerate(rules)}

    for rule in rules:
        rule.is_lrec = False
        rule.is_memo = not rule.no_memo

    edges = [_callable_rule_ids(rule.exp, rule_index) for rule in rules]

    state = [State.FIRST] * len(rules)
    stack: list[int] = []

    def dfs(rule_id: int) -> None:
        if state[rule_id] in {State.VISITING, State.VISITED}:
            return

        state[rule_id] = State.VISITING
        stack.append(rule_id)

        for child_id in edges[rule_id]:
            if state[child_id] == State.FIRST:
                dfs(child_id)
            elif state[child_id] == State.VISITING:
                rules[child_id].is_lrec = True
                rules[child_id].is_memo = False
                for rid in stack:
                    rules[rid].is_memo = False

        stack.pop()
        state[rule_id] = State.VISITED

    for i in range(len(rules)):
        dfs(i)

    return [rule for rule in rules if rule.is_lrec]



# Copyright (c) 2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from pathlib import Path

import pytest

from tatsu.tool.ideps import Dependency, findeps, moduledeps, render


def test_moduledeps_collects_absolute_and_relative_imports(tmp_path: Path) -> None:
    pkg = tmp_path / "pkg"
    pkg.mkdir()
    mod = pkg / "mod.py"
    mod.write_text(
        "from __future__ import annotations\n"
        "import os\n"
        "from collections import defaultdict\n"
        "from . import sibling\n"
        "from .subpkg import child\n"
        "from .. import parent\n"
        "from ..other import thing\n"
    )

    result = moduledeps("pkg.mod", mod)

    assert result.name == "pkg.mod"
    assert result.path == mod

    # `moduledeps` records imported modules/packages, not individual
    # names, and returns them sorted by name.
    assert set(result.imports) == {
        Dependency("collections", is_relative=False),
        Dependency("os", is_relative=False),
        Dependency("parent", is_relative=True),
        Dependency("other", is_relative=True),
        Dependency("pkg.sibling", is_relative=True),
        Dependency("pkg.subpkg", is_relative=True),
    }


def test_findeps_uses_module_names_from_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    pkg = tmp_path / "pkg"
    pkg.mkdir()

    a = pkg / "a.py"
    b = pkg / "b.py"
    a.write_text("import os\n")
    b.write_text("import sys\n")

    monkeypatch.chdir(tmp_path)

    results = findeps([Path("pkg/a.py"), Path("pkg/b.py")])

    assert [m.name for m in results] == ["pkg.a", "pkg.b"]
    # `findeps` preserves the relative paths it receives; when resolved,
    # they point to the files we created under `tmp_path`.
    assert [m.path.resolve() for m in results] == [a, b]


def test_render_builds_tree_with_internal_and_external_deps(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    pkg = tmp_path / "pkg"
    pkg.mkdir()

    a = pkg / "a.py"
    b = pkg / "b.py"
    a.write_text("from __future__ import annotations\nimport os\nfrom . import b\n")
    b.write_text("import sys\n")

    monkeypatch.chdir(tmp_path)

    results = findeps([Path("pkg/a.py"), Path("pkg/b.py")])
    tree = render(results)

    # With a single top-level package, render() returns that node directly.
    assert str(tree.label) == "pkg"

    mod_a = next(child for child in tree.children if str(child.label) == "a")
    labels = {str(child.label) for child in mod_a.children}

    # Internal sibling dependency and external import are both shown.
    assert "○ .b" in labels
    assert "⟨os⟩" in labels



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import pytest


class lisp(list):
    """
    A list subclass designed for boolean-chain parsing.
    Acts as a truthy anchor and a pass-through recorder.
    """

    def __call__(self, item):
        self.append(item)
        # Returns the item to allow the 'and' chain to control flow
        return item

    def __bool__(self) -> bool:
        # Always True so the instance itself doesn't short-circuit the start
        return True


# --- Unit Tests ---


def test_successful_chain():
    """Verifies that a chain of truthy expressions completes and records all."""
    t = lisp()

    # Simulating a successful parse: rule = a b c;
    result = t(a := "match_a") and t(b := "match_b") and t("match_c")

    assert all(t)
    assert len(t) == 3
    assert a == "match_a"
    assert b == "match_b"
    assert t == ["match_a", "match_b", "match_c"]
    assert result == "match_c"  # The last truthy value


def test_short_circuit_on_falsy():
    """Verifies that the chain stops at the first None/False but records it."""
    t = lisp()

    def match_fail():
        return None

    # Simulating: rule = head fail_point never_reached;
    (t(head := "valid_head") and t(fail := match_fail()) and t("never_reached"))

    assert not all(t)
    assert len(t) == 2  # Stopped after the second term
    assert t[-1] is None  # The failure was recorded
    assert head == "valid_head"
    # Note: 'fail' is bound to None in the local scope


def test_nested_subexpression():
    """Verifies that sub-expressions work within the t() call."""
    t = lisp()

    # rule = head (sub1 sub2) tail;
    (t(head := "A") and t(sub := ("B1" and "B2")) and t("C"))

    assert t == ["A", "B2", "C"]
    assert sub == "B2"


def test_empty_list_truthiness():
    """Ensures the worker itself doesn't kill the chain even if empty."""
    t = lisp()
    assert bool(t) is True

    # Even if we append a falsy value first, the 'and' logic handles it
    chain = t(0) and t(1)
    assert list(t) == [0]
    assert chain == 0


if __name__ == "__main__":
    # To run this: pytest lisp_worker.py
    pytest.main([__file__])



import re
from typing import Any


def rle_encode(text: str) -> str:
    """Safely encodes runs of characters by doubling existing delimiters and using ~cN~ for runs."""
    # Step 1: Escape existing literal delimiters by doubling them up (~ -> ~~)
    escaped = text.replace("~", "~~")

    # Step 2: Compress runs of 4 or more identical characters
    # Negative-lookahead ensures we don't accidentally compress our newly doubled '~'
    pattern = re.compile(r"([^~])\1{3,}")
    return pattern.sub(lambda m: f"~{m.group(1)}{len(m.group(0))}~", escaped)


def rle_decode(text: str) -> str:
    """Decodes markers and handles escaped literal delimiters properly."""
    # Step 1: Find and expand the RLE tokens (~cN~)
    # Strictly matches one non-tilde character and its count inside ~ delimiters
    rle_pattern = re.compile(r"~([^~])(\d+)~")
    expanded = rle_pattern.sub(lambda m: m.group(1) * int(m.group(2)), text)

    # Step 2: Collapse the doubled literal delimiters back to single ones (~~ -> ~)
    return expanded.replace("~~", "~")


def compact_value(data: Any) -> Any:
    """Recursively targets string data for bulletproof RLE compression."""
    if isinstance(data, str):
        return rle_encode(data)
    if isinstance(data, dict):
        return {k: compact_value(v) for k, v in data.items()}
    if isinstance(data, list):
        return [compact_value(i) for i in data]
    return data


def decompact_value(data: Any) -> Any:
    """Recursively walks the data structure to expand RLE strings back to original."""
    if isinstance(data, str):
        return rle_decode(data)
    if isinstance(data, dict):
        return {k: decompact_value(v) for k, v in data.items()}
    if isinstance(data, list):
        return [decompact_value(i) for i in data]
    return data



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import random
import threading
import time
from typing import Any

from ..ztyle import Color, Style  # noqa: F401
from .bar import Bar
from .multi import Multi
from .row import BarRow, Metrics


def main() -> None:
    """Visual test — run with ``python -m tatsu.util.barz``."""
    c = Color(True)

    class StyleRow(BarRow):
        def render(self, m: Metrics) -> list[Any]:
            s = Style()
            bar = Bar(
                fill="=>`",
                style=[s.green(), s.green(), s.white().dim()],
            )
            bar.update(m.pos, m.total)
            return [
                f"{s(str(m.label), fmt=">20s")} ",
                bar,
                f"{m.percentage:3.0f}% ",
                # f"{m.elapsed / 1_000_000_000:4.1f}s",
                f"{m.m:02d}:{m.s:02d}.{m.ms:03d}",
            ]

    s = c.style()
    red = s.red()
    green = s.green()
    blue = s.blue().bold()

    overall = BarRow(
        label="overall: ",
        fill="---",
        style=[green, green, s.dim()],
    )
    bars: list[BarRow] = [
        overall,
        StyleRow(label=red("lexing"), style=[red]),
        BarRow(label="parsing"),
        BarRow(label="semantics", total=200),
        StyleRow(label="codegen", total=500, style=[blue]),
        BarRow(label="testing", total=50),
    ]
    overall.update(0, len(bars))

    m = Multi(bars)

    def worker(bar: BarRow, delay: float, step: int, overall: BarRow):
        bar.start()
        m.print(f"starting {bar.label}")
        while bar.pos < bar.total:
            time.sleep(delay)
            bar.update(min(bar.pos + random.randint(1, step), bar.total))  # noqa: S311
        m.print(blue(f"{step} finished {bar.label}"))
        overall.update(overall.pos + 1)

    threads = [
        threading.Thread(target=worker, args=(b, d, s, overall), daemon=True)
        for b, d, s in [
            (bars[1], 0.10, 8),
            (bars[2], 0.16, 5),
            (bars[3], 0.20, 10),
            (bars[4], 0.08, 4),
            (bars[5], 0.30, 2),
        ]
    ]

    m.print(red("workers starting up"))
    m.print(blue("pipeline warming"))

    m.start()

    for b in bars:
        b.start()
    try:
        for t in threads:
            t.start()
        for t in threads:
            t.join()
        time.sleep(2.0)
    except KeyboardInterrupt:
        pass
    m.stop()


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""CLI for TatSu, mirroring ogopego's three-subcommand structure."""

from __future__ import annotations

import argparse
import signal
import sys
from argparse import ArgumentParser

from .. import __toolname__, __version__, g2e
from ..exceptions import ParseException
from ..tool import bench, ideps
from ..util.checkpygments import is_pygments_available
from .boot_cmd import add_boot_cmd, boot_cmd
from .cfg import CLIConfig, CLIError
from .global_opt import add_global_options
from .grammar_cmd import add_grammar_cmd, grammar_cmd
from .out import output_results
from .run_cmd import run_cmd
from .run_opt import add_run_cmd


TITLE = "竜TatSu"
VERSION = f"{TITLE} v{__version__}"
DESCRIPTION = (
    f'{TITLE} takes a grammar in extended EBNF'
    ' as input, and outputs a memoizing'
    ' PEG/Packrat parser in Python.'
)


def create_argument_parser() -> ArgumentParser:

    # Handle --version before argparse to match ogopego's pre-dispatch check
    parser = argparse.ArgumentParser(
        prog="tatsu",
        description=DESCRIPTION,
    )

    sub = parser.add_subparsers(
        dest="command",
        description="Main execution mode",
        required=False,
        help="Available subcommands",
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"{__toolname__} {__version__}",
        help="Print version information and exit",
    )
    add_global_options(parser)
    # _help_cmd = add_help_cmd(subparsers)
    _boot_cmd = add_boot_cmd(sub)
    _grammar_cmd = add_grammar_cmd(sub)
    _run_cmd = add_run_cmd(sub)
    _g2e_cmd = g2e.add_g2e_cmd(sub)
    _bench_cmd = bench.add_bench_cmd(sub)
    _ideps_cmd = ideps.add_ideps_cmd(sub)
    return parser


def run_cling_cli(parser: argparse.ArgumentParser) -> CLIConfig:
    """Parse command-line arguments and return a CLIConfig.

    Matches ogopego's subcommand structure (run / boot / grammar).
    """
    cfg = CLIConfig()
    namespace = parser.parse_args(namespace=cfg)

    for name, value in vars(namespace).items():
        if name == "name":
            continue
        if not hasattr(cfg, name):
            raise RuntimeError(f"Unknown option: {name}")
        setattr(cfg, name, value)
    return cfg


def cling_main() -> int:  # noqa: PLR0911
    """Entry point for the cling CLI (not wired to console_scripts yet)."""

    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    parser = create_argument_parser()
    try:
        cfg = run_cling_cli(parser)

        if cfg.style == "list":
            if is_pygments_available():
                from pygments.styles import get_all_styles

                for style in sorted(get_all_styles()):
                    print(style)
            else:
                print("pygments not installed", file=sys.stderr)
                print("install with: pip install tatsu[cling]", file=sys.stderr)
            return 0
        match cfg.command:
            case "boot":
                results = boot_cmd(cfg)
            case "grammar":
                results = grammar_cmd(cfg)
            case "run":
                results = run_cmd(cfg)
            case "bench":
                return bench.bench_cmd(parser)
            case "g2e":
                return g2e.g2e_cmd(parser)
            case "ideps":
                return ideps.ideps_cmd(parser)
            case _:
                parser.print_usage()
                return 1
        output_results(cfg, results)
        return 0

    except CLIError as e:
        print(e, file=sys.stderr)
        return 1
    except ParseException as e:
        print(e, file=sys.stderr)
        return 1
    except BrokenPipeError:
        return signal.SIGPIPE + 0o200
    except KeyboardInterrupt:
        return signal.SIGINT + 0o200


def add_help_cmd(subparsers):
    help_parser = subparsers.add_parser(
        "help",
        help="Provide command help",
        add_help=False,
    )
    return help_parser



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .cursor import Cursor, Text, Tokenizer
from .infos import LineIndexInfo, LineInfo, PosLine
from .null import NullCursor, NullText


__all__ = [
    'Cursor',
    'NullCursor',
    'Text',
    'NullText',
    'LineIndexInfo',
    'LineInfo',
    'PosLine',
    'Tokenizer',
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# by Big Pickle 2026-06-21
from __future__ import annotations

from typing import Any


__all__ = ["Tree"]


class Tree:
    def __init__(self, label: Any, children: list[Tree] | None = None):
        self.label = label
        self.children: list[Tree] = children or []

    def add(self, label: Any) -> Tree:
        child = Tree(label)
        return self.add_child(child)

    def add_child(self, child: Tree) -> Tree:
        self.children.append(child)
        return child

    def __str__(self) -> str:
        return self.render()

    def render(self) -> str:
        lines = [f"{self.label}"]
        if not self.children:
            return "\n".join(lines)

        for child in self.children[:-1]:
            sub = child.render().splitlines()
            lines.append("├── " + sub[0])
            for line in sub[1:]:
                lines += ["│   " + line]

        child = self.children[-1]
        sub = child.render().splitlines()
        lines.append("└── " + sub[0])
        for line in sub[1:]:
            lines += ["    " + line]

        return "\n".join(lines)



from dataclasses import dataclass, field


@dataclass(slots=True, kw_only=True)
class Bar:
    name: str
    total: int = 100
    current: int = 0
    start_time: float = field(default_factory=time.time)

    def advance(self, delta: int = 1):
        self.current = min(self.current + delta, self.total)

    def call_render(self, width: int) -> list[Style]:
        """
        The private bridge execution method. Runs completely on the background
        thread time. Computes all fields before calling render().
        """
        now = time.time()
        elapsed = now - self.start_time
        pct = (self.current / self.total * 100) if self.total > 0 else 0.0
        rate = self.current / elapsed if elapsed > 0 else 0
        eta = (self.total - self.current) / rate if rate > 0 else 0.0

        return self.render(
            width=width,
            pos=self.current,
            total=self.total,
            pct=pct,
            elapsed=elapsed,
            eta=eta,
        )

    def render(self, *, width: int, pct: float, eta: float, **kwargs) -> list[Style]:
        """
        The default user-overridable hook. Simple, predictable, and
        gives the user exactly the computed values they want.
        """
        color = Color.default()
        prefix = f"{self.name:<20} "
        suffix = f" {pct:>3.0f}% (ETA: {eta:.1f}s)"

        fixed_len = len(prefix) + len(suffix) + 2
        bar_budget = max(0, width - fixed_len)

        fill_len = int(bar_budget * (pct / 100))
        empty_len = bar_budget - fill_len
        blocks_str = f"[{'█' * fill_len}{'-' * empty_len}]"

        return [
            color.style(prefix, bold=True),
            color.style(blocks_str, fg=8),
            color.style(suffix, fg=2),
        ]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Self

from .style import Style


class XStyle(Style):
    def acid_green(self) -> Self:
        return self.fg(148)

    def acid_green_bg(self) -> Self:
        return self.bg(148)

    def amethyst(self) -> Self:
        return self.fg(99)

    def amethyst_bg(self) -> Self:
        return self.bg(99)

    def aquamarine_2(self) -> Self:
        return self.fg(86)

    def aquamarine_2_bg(self) -> Self:
        return self.bg(86)

    def aquamarine_3(self) -> Self:
        return self.fg(80)

    def aquamarine_3_bg(self) -> Self:
        return self.bg(80)

    def banana(self) -> Self:
        return self.fg(228)

    def banana_bg(self) -> Self:
        return self.bg(228)

    def cube_black(self) -> Self:
        return self.fg(16)

    def blue256(self) -> Self:
        return self.fg(21)

    def blue_bright(self) -> Self:
        return self.fg(20)

    def blue_bright_bg(self) -> Self:
        return self.bg(20)

    def blue_violet(self) -> Self:
        return self.fg(56)

    def blue_violet_bg(self) -> Self:
        return self.bg(56)

    def cube_bright_black(self) -> Self:
        return self.fg(8)

    def cube_bright_blue(self) -> Self:
        return self.fg(12)

    def cube_bright_cyan(self) -> Self:
        return self.fg(14)

    def cube_bright_green(self) -> Self:
        return self.fg(10)

    def cube_bright_magenta(self) -> Self:
        return self.fg(13)

    def cube_bright_red(self) -> Self:
        return self.fg(154)

    def cube_bright_white(self) -> Self:
        return self.fg(15)

    def cube_bright_yellow(self) -> Self:
        return self.fg(11)

    def brown(self) -> Self:
        return self.fg(94)

    def brown_bg(self) -> Self:
        return self.bg(94)

    def cadet_blue(self) -> Self:
        return self.fg(66)

    def cadet_blue_bg(self) -> Self:
        return self.bg(66)

    def chartreuse_1(self) -> Self:
        return self.fg(118)

    def chartreuse_1_bg(self) -> Self:
        return self.bg(118)

    def chartreuse_2(self) -> Self:
        return self.fg(112)

    def chartreuse_2_bg(self) -> Self:
        return self.bg(112)

    def chartreuse_3(self) -> Self:
        return self.fg(82)

    def chartreuse_3_bg(self) -> Self:
        return self.bg(82)

    def chartreuse_4(self) -> Self:
        return self.fg(76)

    def chartreuse_4_bg(self) -> Self:
        return self.bg(76)

    def chartreuse_5(self) -> Self:
        return self.fg(70)

    def chartreuse_5_bg(self) -> Self:
        return self.bg(70)

    def chartreuse_6(self) -> Self:
        return self.fg(64)

    def chartreuse_6_bg(self) -> Self:
        return self.bg(64)

    def clay(self) -> Self:
        return self.fg(138)

    def clay_bg(self) -> Self:
        return self.bg(138)

    def coral(self) -> Self:
        return self.fg(203)

    def coral_bg(self) -> Self:
        return self.bg(203)

    def cornflower_blue(self) -> Self:
        return self.fg(68)

    def cornflower_blue_bg(self) -> Self:
        return self.bg(68)

    def cornflower_blue_bright(self) -> Self:
        return self.fg(75)

    def cornflower_blue_bright_bg(self) -> Self:
        return self.bg(75)

    def cotton_candy(self) -> Self:
        return self.fg(218)

    def cotton_candy_bg(self) -> Self:
        return self.bg(218)

    def cream(self) -> Self:
        return self.fg(229)

    def cream_bg(self) -> Self:
        return self.bg(229)

    def crimson(self) -> Self:
        return self.fg(162)

    def crimson_bg(self) -> Self:
        return self.bg(162)

    def cyan256(self) -> Self:
        return self.fg(51)

    def cyan_bright(self) -> Self:
        return self.fg(45)

    def cyan_bright_bg(self) -> Self:
        return self.bg(45)

    def cyan_luminous(self) -> Self:
        return self.fg(50)

    def cyan_luminous_bg(self) -> Self:
        return self.bg(50)

    def cyan_medium(self) -> Self:
        return self.fg(43)

    def cyan_medium_bg(self) -> Self:
        return self.bg(43)

    def dark_blue(self) -> Self:
        return self.fg(18)

    def dark_blue_bg(self) -> Self:
        return self.bg(18)

    def dark_cyan(self) -> Self:
        return self.fg(36)

    def dark_cyan_bg(self) -> Self:
        return self.bg(36)

    def dark_green(self) -> Self:
        return self.fg(22)

    def dark_green_bg(self) -> Self:
        return self.bg(22)

    def dark_magenta(self) -> Self:
        return self.fg(53)

    def dark_magenta_bg(self) -> Self:
        return self.bg(53)

    def dark_orange(self) -> Self:
        return self.fg(208)

    def dark_orange_bg(self) -> Self:
        return self.bg(208)

    def dark_orange_3(self) -> Self:
        return self.fg(166)

    def dark_orange_3_bg(self) -> Self:
        return self.bg(166)

    def dark_red(self) -> Self:
        return self.fg(88)

    def dark_red_bg(self) -> Self:
        return self.bg(88)

    def dark_sea_green_1(self) -> Self:
        return self.fg(108)

    def dark_sea_green_1_bg(self) -> Self:
        return self.bg(108)

    def dark_sea_green_2(self) -> Self:
        return self.fg(107)

    def dark_sea_green_2_bg(self) -> Self:
        return self.bg(107)

    def dark_sea_green_3(self) -> Self:
        return self.fg(77)

    def dark_sea_green_3_bg(self) -> Self:
        return self.bg(77)

    def dark_sea_green_4(self) -> Self:
        return self.fg(71)

    def dark_sea_green_4_bg(self) -> Self:
        return self.bg(71)

    def dark_sea_green_5(self) -> Self:
        return self.fg(65)

    def dark_sea_green_5_bg(self) -> Self:
        return self.bg(65)

    def dark_sea_green_bright(self) -> Self:
        return self.fg(114)

    def dark_sea_green_bright_bg(self) -> Self:
        return self.bg(114)

    def dark_slate_gray_bright(self) -> Self:
        return self.fg(87)

    def dark_slate_gray_bright_bg(self) -> Self:
        return self.bg(87)

    def dark_turquoise(self) -> Self:
        return self.fg(44)

    def dark_turquoise_bg(self) -> Self:
        return self.bg(44)

    def dark_violet(self) -> Self:
        return self.fg(129)

    def dark_violet_bg(self) -> Self:
        return self.bg(129)

    def deep_pink_1(self) -> Self:
        return self.fg(156)

    def deep_pink_1_bg(self) -> Self:
        return self.bg(156)

    def deep_pink_2(self) -> Self:
        return self.fg(155)

    def deep_pink_2_bg(self) -> Self:
        return self.bg(155)

    def deep_pink_3(self) -> Self:
        return self.fg(125)

    def deep_pink_3_bg(self) -> Self:
        return self.bg(125)

    def deep_pink_4(self) -> Self:
        return self.fg(89)

    def deep_pink_4_bg(self) -> Self:
        return self.bg(89)

    def deep_pink_bright(self) -> Self:
        return self.fg(197)

    def deep_pink_bright_bg(self) -> Self:
        return self.bg(197)

    def deep_sky_blue_2(self) -> Self:
        return self.fg(39)

    def deep_sky_blue_2_bg(self) -> Self:
        return self.bg(39)

    def deep_sky_blue_3(self) -> Self:
        return self.fg(38)

    def deep_sky_blue_3_bg(self) -> Self:
        return self.bg(38)

    def deep_sky_blue_4(self) -> Self:
        return self.fg(32)

    def deep_sky_blue_4_bg(self) -> Self:
        return self.bg(32)

    def deep_sky_blue_5(self) -> Self:
        return self.fg(31)

    def deep_sky_blue_5_bg(self) -> Self:
        return self.bg(31)

    def deep_sky_blue_6(self) -> Self:
        return self.fg(25)

    def deep_sky_blue_6_bg(self) -> Self:
        return self.bg(25)

    def deep_sky_blue_7(self) -> Self:
        return self.fg(24)

    def deep_sky_blue_7_bg(self) -> Self:
        return self.bg(24)

    def dodger_blue_1(self) -> Self:
        return self.fg(33)

    def dodger_blue_1_bg(self) -> Self:
        return self.bg(33)

    def dodger_blue_2(self) -> Self:
        return self.fg(27)

    def dodger_blue_2_bg(self) -> Self:
        return self.bg(27)

    def dodger_blue_3(self) -> Self:
        return self.fg(26)

    def dodger_blue_3_bg(self) -> Self:
        return self.bg(26)

    def electric_indigo(self) -> Self:
        return self.fg(57)

    def electric_indigo_bg(self) -> Self:
        return self.bg(57)

    def fuchsia(self) -> Self:
        return self.fg(199)

    def fuchsia_bg(self) -> Self:
        return self.bg(199)

    def gold(self) -> Self:
        return self.fg(220)

    def gold_bg(self) -> Self:
        return self.bg(220)

    def gold_dark(self) -> Self:
        return self.fg(136)

    def gold_dark_bg(self) -> Self:
        return self.bg(136)

    def gold_medium(self) -> Self:
        return self.fg(178)

    def gold_medium_bg(self) -> Self:
        return self.bg(178)

    def gray003(self) -> Self:
        return self.fg(232)

    def gray003_bg(self) -> Self:
        return self.bg(232)

    def gray007(self) -> Self:
        return self.fg(233)

    def gray007_bg(self) -> Self:
        return self.bg(233)

    def gray011(self) -> Self:
        return self.fg(234)

    def gray011_bg(self) -> Self:
        return self.bg(234)

    def gray015(self) -> Self:
        return self.fg(235)

    def gray015_bg(self) -> Self:
        return self.bg(235)

    def gray019(self) -> Self:
        return self.fg(236)

    def gray019_bg(self) -> Self:
        return self.bg(236)

    def gray023(self) -> Self:
        return self.fg(237)

    def gray023_bg(self) -> Self:
        return self.bg(237)

    def gray027(self) -> Self:
        return self.fg(238)

    def gray027_bg(self) -> Self:
        return self.bg(238)

    def gray031(self) -> Self:
        return self.fg(239)

    def gray031_bg(self) -> Self:
        return self.bg(239)

    def gray035(self) -> Self:
        return self.fg(240)

    def gray035_bg(self) -> Self:
        return self.bg(240)

    def gray039(self) -> Self:
        return self.fg(241)

    def gray039_bg(self) -> Self:
        return self.bg(241)

    def gray043(self) -> Self:
        return self.fg(242)

    def gray043_bg(self) -> Self:
        return self.bg(242)

    def gray047(self) -> Self:
        return self.fg(243)

    def gray047_bg(self) -> Self:
        return self.bg(243)

    def gray051(self) -> Self:
        return self.fg(244)

    def gray051_bg(self) -> Self:
        return self.bg(244)

    def gray055(self) -> Self:
        return self.fg(245)

    def gray055_bg(self) -> Self:
        return self.bg(245)

    def gray059(self) -> Self:
        return self.fg(246)

    def gray059_bg(self) -> Self:
        return self.bg(246)

    def gray063(self) -> Self:
        return self.fg(247)

    def gray063_bg(self) -> Self:
        return self.bg(247)

    def gray067(self) -> Self:
        return self.fg(248)

    def gray067_bg(self) -> Self:
        return self.bg(248)

    def gray071(self) -> Self:
        return self.fg(249)

    def gray071_bg(self) -> Self:
        return self.bg(249)

    def gray075(self) -> Self:
        return self.fg(250)

    def gray075_bg(self) -> Self:
        return self.bg(250)

    def gray079(self) -> Self:
        return self.fg(251)

    def gray079_bg(self) -> Self:
        return self.bg(251)

    def gray083(self) -> Self:
        return self.fg(252)

    def gray083_bg(self) -> Self:
        return self.bg(252)

    def gray087(self) -> Self:
        return self.fg(253)

    def gray087_bg(self) -> Self:
        return self.bg(253)

    def gray091(self) -> Self:
        return self.fg(254)

    def gray091_bg(self) -> Self:
        return self.bg(254)

    def gray095(self) -> Self:
        return self.fg(255)

    def gray095_bg(self) -> Self:
        return self.bg(255)

    def green256(self) -> Self:
        return self.fg(2)

    def green_bright(self) -> Self:
        return self.fg(40)

    def green_bright_bg(self) -> Self:
        return self.bg(40)

    def green_forest(self) -> Self:
        return self.fg(28)

    def green_forest_bg(self) -> Self:
        return self.bg(28)

    def green_yellow(self) -> Self:
        return self.fg(185)

    def green_yellow_bg(self) -> Self:
        return self.bg(185)

    def grey_37(self) -> Self:
        return self.fg(59)

    def grey_37_bg(self) -> Self:
        return self.bg(59)

    def grey_46(self) -> Self:
        return self.fg(96)

    def grey_46_bg(self) -> Self:
        return self.bg(96)

    def grey_53(self) -> Self:
        return self.fg(101)

    def grey_53_bg(self) -> Self:
        return self.bg(101)

    def grey_53_neutral(self) -> Self:
        return self.fg(102)

    def grey_53_neutral_bg(self) -> Self:
        return self.bg(102)

    def grey_69(self) -> Self:
        return self.fg(139)

    def grey_69_bg(self) -> Self:
        return self.bg(139)

    def grey_74(self) -> Self:
        return self.fg(145)

    def grey_74_bg(self) -> Self:
        return self.bg(145)

    def grey_84(self) -> Self:
        return self.fg(181)

    def grey_84_bg(self) -> Self:
        return self.bg(181)

    def grey_84_neutral(self) -> Self:
        return self.fg(182)

    def grey_84_neutral_bg(self) -> Self:
        return self.bg(182)

    def honeydew(self) -> Self:
        return self.fg(188)

    def honeydew_bg(self) -> Self:
        return self.bg(188)

    def hot_pink(self) -> Self:
        return self.fg(204)

    def hot_pink_bg(self) -> Self:
        return self.bg(204)

    def hot_pink_bright(self) -> Self:
        return self.fg(205)

    def hot_pink_bright_bg(self) -> Self:
        return self.bg(205)

    def indian_red(self) -> Self:
        return self.fg(161)

    def indian_red_bg(self) -> Self:
        return self.bg(161)

    def ivory(self) -> Self:
        return self.fg(230)

    def ivory_bg(self) -> Self:
        return self.bg(230)

    def khaki_1(self) -> Self:
        return self.fg(144)

    def khaki_1_bg(self) -> Self:
        return self.bg(144)

    def khaki_2(self) -> Self:
        return self.fg(143)

    def khaki_2_bg(self) -> Self:
        return self.bg(143)

    def khaki_3(self) -> Self:
        return self.fg(142)

    def khaki_3_bg(self) -> Self:
        return self.bg(142)

    def khaki_4(self) -> Self:
        return self.fg(100)

    def khaki_4_bg(self) -> Self:
        return self.bg(100)

    def khaki_bright(self) -> Self:
        return self.fg(179)

    def khaki_bright_bg(self) -> Self:
        return self.bg(179)

    def lavender(self) -> Self:
        return self.fg(183)

    def lavender_bg(self) -> Self:
        return self.bg(183)

    def lavender_blue(self) -> Self:
        return self.fg(147)

    def lavender_blue_bg(self) -> Self:
        return self.bg(147)

    def lavender_purple(self) -> Self:
        return self.fg(141)

    def lavender_purple_bg(self) -> Self:
        return self.bg(141)

    def light_blue_sky(self) -> Self:
        return self.fg(153)

    def light_blue_sky_bg(self) -> Self:
        return self.bg(153)

    def light_coral(self) -> Self:
        return self.fg(210)

    def light_coral_bg(self) -> Self:
        return self.bg(210)

    def light_cyan_2(self) -> Self:
        return self.fg(189)

    def light_cyan_2_bg(self) -> Self:
        return self.bg(189)

    def light_cyan_3(self) -> Self:
        return self.fg(116)

    def light_cyan_3_bg(self) -> Self:
        return self.bg(116)

    def light_green_1(self) -> Self:
        return self.fg(119)

    def light_green_1_bg(self) -> Self:
        return self.bg(119)

    def light_green_2(self) -> Self:
        return self.fg(84)

    def light_green_2_bg(self) -> Self:
        return self.bg(84)

    def light_green_3(self) -> Self:
        return self.fg(83)

    def light_green_3_bg(self) -> Self:
        return self.bg(83)

    def light_grey_blue(self) -> Self:
        return self.fg(146)

    def light_grey_blue_bg(self) -> Self:
        return self.bg(146)

    def light_khaki(self) -> Self:
        return self.fg(186)

    def light_khaki_bg(self) -> Self:
        return self.bg(186)

    def light_orchid(self) -> Self:
        return self.fg(212)

    def light_orchid_bg(self) -> Self:
        return self.bg(212)

    def light_pink(self) -> Self:
        return self.fg(169)

    def light_pink_bg(self) -> Self:
        return self.bg(169)

    def light_salmon(self) -> Self:
        return self.fg(168)

    def light_salmon_bg(self) -> Self:
        return self.bg(168)

    def light_salmon_2(self) -> Self:
        return self.fg(174)

    def light_salmon_2_bg(self) -> Self:
        return self.bg(174)

    def light_salmon_3(self) -> Self:
        return self.fg(216)

    def light_salmon_3_bg(self) -> Self:
        return self.bg(216)

    def light_sea_green(self) -> Self:
        return self.fg(37)

    def light_sea_green_bg(self) -> Self:
        return self.bg(37)

    def light_sky_blue_3(self) -> Self:
        return self.fg(110)

    def light_sky_blue_3_bg(self) -> Self:
        return self.bg(110)

    def light_sky_blue_4(self) -> Self:
        return self.fg(109)

    def light_sky_blue_4_bg(self) -> Self:
        return self.bg(109)

    def light_slate_grey(self) -> Self:
        return self.fg(103)

    def light_slate_grey_bg(self) -> Self:
        return self.bg(103)

    def light_yellow_2(self) -> Self:
        return self.fg(227)

    def light_yellow_2_bg(self) -> Self:
        return self.bg(227)

    def lime(self) -> Self:
        return self.fg(46)

    def lime_bg(self) -> Self:
        return self.bg(46)

    def lime_green(self) -> Self:
        return self.fg(34)

    def lime_green_bg(self) -> Self:
        return self.bg(34)

    def magenta256(self) -> Self:
        return self.fg(201)

    def magenta_bright(self) -> Self:
        return self.fg(128)

    def magenta_bright_bg(self) -> Self:
        return self.bg(128)

    def magenta_dark(self) -> Self:
        return self.fg(127)

    def magenta_dark_bg(self) -> Self:
        return self.bg(127)

    def magenta_light(self) -> Self:
        return self.fg(207)

    def magenta_light_bg(self) -> Self:
        return self.bg(207)

    def magenta_luminous(self) -> Self:
        return self.fg(157)

    def magenta_luminous_bg(self) -> Self:
        return self.bg(157)

    def magenta_medium(self) -> Self:
        return self.fg(126)

    def magenta_medium_bg(self) -> Self:
        return self.bg(126)

    def magenta_neon(self) -> Self:
        return self.fg(159)

    def magenta_neon_bg(self) -> Self:
        return self.bg(159)

    def magenta_rose(self) -> Self:
        return self.fg(198)

    def magenta_rose_bg(self) -> Self:
        return self.bg(198)

    def magenta_vibrant(self) -> Self:
        return self.fg(158)

    def magenta_vibrant_bg(self) -> Self:
        return self.bg(158)

    def maroon(self) -> Self:
        return self.fg(52)

    def maroon_bg(self) -> Self:
        return self.bg(52)

    def medium_aquamarine(self) -> Self:
        return self.fg(72)

    def medium_aquamarine_bg(self) -> Self:
        return self.bg(72)

    def medium_aquamarine_bright(self) -> Self:
        return self.fg(79)

    def medium_aquamarine_bright_bg(self) -> Self:
        return self.bg(79)

    def medium_blue(self) -> Self:
        return self.fg(19)

    def medium_blue_bg(self) -> Self:
        return self.bg(19)

    def medium_orchid(self) -> Self:
        return self.fg(165)

    def medium_orchid_bg(self) -> Self:
        return self.bg(165)

    def medium_purple_1(self) -> Self:
        return self.fg(135)

    def medium_purple_1_bg(self) -> Self:
        return self.bg(135)

    def medium_purple_2(self) -> Self:
        return self.fg(134)

    def medium_purple_2_bg(self) -> Self:
        return self.bg(134)

    def medium_purple_3(self) -> Self:
        return self.fg(133)

    def medium_purple_3_bg(self) -> Self:
        return self.bg(133)

    def medium_purple_4(self) -> Self:
        return self.fg(104)

    def medium_purple_4_bg(self) -> Self:
        return self.bg(104)

    def medium_purple_5(self) -> Self:
        return self.fg(98)

    def medium_purple_5_bg(self) -> Self:
        return self.bg(98)

    def medium_purple_6(self) -> Self:
        return self.fg(97)

    def medium_purple_6_bg(self) -> Self:
        return self.bg(97)

    def medium_purple_7(self) -> Self:
        return self.fg(60)

    def medium_purple_7_bg(self) -> Self:
        return self.bg(60)

    def medium_red(self) -> Self:
        return self.fg(124)

    def medium_red_bg(self) -> Self:
        return self.bg(124)

    def medium_spring_green(self) -> Self:
        return self.fg(49)

    def medium_spring_green_bg(self) -> Self:
        return self.bg(49)

    def medium_turquoise(self) -> Self:
        return self.fg(73)

    def medium_turquoise_bg(self) -> Self:
        return self.bg(73)

    def mint_green(self) -> Self:
        return self.fg(151)

    def mint_green_bg(self) -> Self:
        return self.bg(151)

    def misty_rose(self) -> Self:
        return self.fg(224)

    def misty_rose_bg(self) -> Self:
        return self.bg(224)

    def navajo_white(self) -> Self:
        return self.fg(221)

    def navajo_white_bg(self) -> Self:
        return self.bg(221)

    def navy(self) -> Self:
        return self.fg(17)

    def navy_bg(self) -> Self:
        return self.bg(17)

    def olive_dark(self) -> Self:
        return self.fg(58)

    def olive_dark_bg(self) -> Self:
        return self.bg(58)

    def olive_drab_3(self) -> Self:
        return self.fg(113)

    def olive_drab_3_bg(self) -> Self:
        return self.bg(113)

    def orange(self) -> Self:
        return self.fg(214)

    def orange_bg(self) -> Self:
        return self.bg(214)

    def orange_darker(self) -> Self:
        return self.fg(160)

    def orange_darker_bg(self) -> Self:
        return self.bg(160)

    def orange_red(self) -> Self:
        return self.fg(202)

    def orange_red_bg(self) -> Self:
        return self.bg(202)

    def orange_yellow(self) -> Self:
        return self.fg(172)

    def orange_yellow_bg(self) -> Self:
        return self.bg(172)

    def orchid(self) -> Self:
        return self.fg(164)

    def orchid_bg(self) -> Self:
        return self.bg(164)

    def orchid_bright(self) -> Self:
        return self.fg(206)

    def orchid_bright_bg(self) -> Self:
        return self.bg(206)

    def orchid_light(self) -> Self:
        return self.fg(170)

    def orchid_light_bg(self) -> Self:
        return self.bg(170)

    def pale_green_1(self) -> Self:
        return self.fg(115)

    def pale_green_1_bg(self) -> Self:
        return self.bg(115)

    def pale_green_2(self) -> Self:
        return self.fg(85)

    def pale_green_2_bg(self) -> Self:
        return self.bg(85)

    def pale_green_3(self) -> Self:
        return self.fg(78)

    def pale_green_3_bg(self) -> Self:
        return self.bg(78)

    def pale_green_bright(self) -> Self:
        return self.fg(120)

    def pale_green_bright_bg(self) -> Self:
        return self.bg(120)

    def pale_turquoise_2(self) -> Self:
        return self.fg(123)

    def pale_turquoise_2_bg(self) -> Self:
        return self.bg(123)

    def pale_turquoise_3(self) -> Self:
        return self.fg(122)

    def pale_turquoise_3_bg(self) -> Self:
        return self.bg(122)

    def pale_violet_red(self) -> Self:
        return self.fg(132)

    def pale_violet_red_bg(self) -> Self:
        return self.bg(132)

    def peach_puff(self) -> Self:
        return self.fg(222)

    def peach_puff_bg(self) -> Self:
        return self.bg(222)

    def peru(self) -> Self:
        return self.fg(167)

    def peru_bg(self) -> Self:
        return self.bg(167)

    def pink(self) -> Self:
        return self.fg(200)

    def pink_bg(self) -> Self:
        return self.bg(200)

    def pink_light(self) -> Self:
        return self.fg(217)

    def pink_light_bg(self) -> Self:
        return self.bg(217)

    def plum_3(self) -> Self:
        return self.fg(163)

    def plum_3_bg(self) -> Self:
        return self.bg(163)

    def plume(self) -> Self:
        return self.fg(177)

    def plume_bg(self) -> Self:
        return self.bg(177)

    def powder_blue(self) -> Self:
        return self.fg(152)

    def powder_blue_bg(self) -> Self:
        return self.bg(152)

    def purple_3(self) -> Self:
        return self.fg(55)

    def purple_3_bg(self) -> Self:
        return self.bg(55)

    def purple_4(self) -> Self:
        return self.fg(54)

    def purple_4_bg(self) -> Self:
        return self.bg(54)

    def purple_bright(self) -> Self:
        return self.fg(92)

    def purple_bright_bg(self) -> Self:
        return self.bg(92)

    def purple_medium(self) -> Self:
        return self.fg(90)

    def purple_medium_bg(self) -> Self:
        return self.bg(90)

    def purple_medium_bright(self) -> Self:
        return self.fg(91)

    def purple_medium_bright_bg(self) -> Self:
        return self.bg(91)

    def redw(self) -> Self:
        return self.fg(196)

    def rosy_brown(self) -> Self:
        return self.fg(95)

    def rosy_brown_bg(self) -> Self:
        return self.bg(95)

    def rosy_pink(self) -> Self:
        return self.fg(175)

    def rosy_pink_bg(self) -> Self:
        return self.bg(175)

    def royal_blue_1(self) -> Self:
        return self.fg(69)

    def royal_blue_1_bg(self) -> Self:
        return self.bg(69)

    def royal_blue_2(self) -> Self:
        return self.fg(63)

    def royal_blue_2_bg(self) -> Self:
        return self.bg(63)

    def salmon(self) -> Self:
        return self.fg(209)

    def salmon_bg(self) -> Self:
        return self.bg(209)

    def sand(self) -> Self:
        return self.fg(215)

    def sand_bg(self) -> Self:
        return self.bg(215)

    def sandy_brown(self) -> Self:
        return self.fg(137)

    def sandy_brown_bg(self) -> Self:
        return self.bg(137)

    def sandy_brown_bright(self) -> Self:
        return self.fg(173)

    def sandy_brown_bright_bg(self) -> Self:
        return self.bg(173)

    def sienna(self) -> Self:
        return self.fg(130)

    def sienna_bg(self) -> Self:
        return self.bg(130)

    def sky_blue_bright(self) -> Self:
        return self.fg(81)

    def sky_blue_bright_bg(self) -> Self:
        return self.bg(81)

    def sky_blue_light(self) -> Self:
        return self.fg(117)

    def sky_blue_light_bg(self) -> Self:
        return self.bg(117)

    def sky_blue_medium(self) -> Self:
        return self.fg(111)

    def sky_blue_medium_bg(self) -> Self:
        return self.bg(111)

    def slate_blue_1(self) -> Self:
        return self.fg(105)

    def slate_blue_1_bg(self) -> Self:
        return self.bg(105)

    def slate_blue_2(self) -> Self:
        return self.fg(62)

    def slate_blue_2_bg(self) -> Self:
        return self.bg(62)

    def slate_blue_3(self) -> Self:
        return self.fg(61)

    def slate_blue_3_bg(self) -> Self:
        return self.bg(61)

    def soft_green(self) -> Self:
        return self.fg(150)

    def soft_green_bg(self) -> Self:
        return self.bg(150)

    def soft_pink(self) -> Self:
        return self.fg(211)

    def soft_pink_bg(self) -> Self:
        return self.bg(211)

    def soft_yellow(self) -> Self:
        return self.fg(187)

    def soft_yellow_bg(self) -> Self:
        return self.bg(187)

    def spindrift(self) -> Self:
        return self.fg(121)

    def spindrift_bg(self) -> Self:
        return self.bg(121)

    def spring_green_1(self) -> Self:
        return self.fg(48)

    def spring_green_1_bg(self) -> Self:
        return self.bg(48)

    def spring_green_2(self) -> Self:
        return self.fg(47)

    def spring_green_2_bg(self) -> Self:
        return self.bg(47)

    def spring_green_3(self) -> Self:
        return self.fg(42)

    def spring_green_3_bg(self) -> Self:
        return self.bg(42)

    def spring_green_4(self) -> Self:
        return self.fg(41)

    def spring_green_4_bg(self) -> Self:
        return self.bg(41)

    def spring_green_5(self) -> Self:
        return self.fg(35)

    def spring_green_5_bg(self) -> Self:
        return self.bg(35)

    def spring_green_6(self) -> Self:
        return self.fg(29)

    def spring_green_6_bg(self) -> Self:
        return self.bg(29)

    def steel_blue_2(self) -> Self:
        return self.fg(74)

    def steel_blue_2_bg(self) -> Self:
        return self.bg(74)

    def steel_blue_3(self) -> Self:
        return self.fg(67)

    def steel_blue_3_bg(self) -> Self:
        return self.bg(67)

    def teal(self) -> Self:
        return self.fg(23)

    def teal_bg(self) -> Self:
        return self.bg(23)

    def terracotta(self) -> Self:
        return self.fg(131)

    def terracotta_bg(self) -> Self:
        return self.bg(131)

    def thistle_1(self) -> Self:
        return self.fg(219)

    def thistle_1_bg(self) -> Self:
        return self.bg(219)

    def thistle_2(self) -> Self:
        return self.fg(176)

    def thistle_2_bg(self) -> Self:
        return self.bg(176)

    def thistle_3(self) -> Self:
        return self.fg(140)

    def thistle_3_bg(self) -> Self:
        return self.bg(140)

    def topaz(self) -> Self:
        return self.fg(223)

    def topaz_bg(self) -> Self:
        return self.bg(223)

    def turquoise_4(self) -> Self:
        return self.fg(30)

    def turquoise_4_bg(self) -> Self:
        return self.bg(30)

    def violet_bright(self) -> Self:
        return self.fg(93)

    def violet_bright_bg(self) -> Self:
        return self.bg(93)

    def violet_light(self) -> Self:
        return self.fg(213)

    def violet_light_bg(self) -> Self:
        return self.bg(213)

    def violet_medium(self) -> Self:
        return self.fg(171)

    def violet_medium_bg(self) -> Self:
        return self.bg(171)

    def wheat(self) -> Self:
        return self.fg(180)

    def wheat_bg(self) -> Self:
        return self.bg(180)

    def cube_white(self) -> Self:
        return self.fg(231)

    def white_pinkish(self) -> Self:
        return self.fg(225)

    def white_pinkish_bg(self) -> Self:
        return self.bg(225)

    def yellow256(self) -> Self:
        return self.fg(226)

    def yellow_green(self) -> Self:
        return self.fg(106)

    def yellow_green_bg(self) -> Self:
        return self.bg(106)

    def yellow_green_bright(self) -> Self:
        return self.fg(149)

    def yellow_green_bright_bg(self) -> Self:
        return self.bg(149)

    def yellow_neon(self) -> Self:
        return self.fg(184)

    def yellow_neon_bg(self) -> Self:
        return self.bg(184)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""
The Buffer class provides the functionality required by a parser-driven lexer.

Line analysis and caching are done so the parser can freely move with goto(parser)
to any position in the parsed text, and still recover accurate information
about source lines and content.
"""

from __future__ import annotations

import re
from functools import cached_property
from pathlib import Path
from typing import Any, Self

from ..config import ParserConfig
from ..util import (
    Undefined,
    cached_re_compile,
    linecount,
    notnone,
    str_from_match,
    typename,
)
from ..util.newlines import take_linebreak_len, take_non_newline_whitespace_len
from ..util.undefined import UndefinedType
from . import LineInfo
from .cursor import Cursor, Text, matchbool, matchfloat, matchint, matchname, matchuint
from .infos import LineIndexInfo, PosLine


DEFAULT_WHITESPACE_RE = re.compile(r'(?m)\s+')

# for backwards compatibility with existing parsers
LineIndexEntry = LineIndexInfo


class BufferCursor(Cursor):
    __slots__ = ('buffer', 'len', 'pos', 'textstr')

    def __init__(self, buffer: Buffer, pos: int = 0):
        super().__init__()
        self.buffer = buffer
        self.textstr = self.buffer.text
        self.pos = pos
        self.len = buffer.len

    def clone(self) -> Self:
        return type(self)(self.buffer, pos=self.pos)

    @property
    def input(self) -> Text:
        return self.buffer

    @property
    def line(self) -> int:
        return self.buffer.posline(self.pos)

    @property
    def linecount(self) -> int:
        return self.buffer.linecount

    @property
    def col(self) -> int:
        return self.buffer.poscol(self.pos)

    @property
    def source(self) -> str:
        n = min(len(self.buffer.lineindex) - 1, self.line)
        source, _line = self.buffer.lineindex[n]
        return source

    @cached_property
    def namechars(self) -> set[str]:
        return self.buffer._namechar_set

    def goto(self, pos: int):
        self.pos = max(0, min(self.buffer.len, pos))

    def move(self, n: int):
        self.goto(self.pos + n)

    def atend(self) -> bool:
        return self.pos >= self.len

    def ateol(self) -> bool:
        return self.atend() or self.current in {'\r', '\n', None}

    @property
    def current(self) -> str | None:
        if self.pos >= self.len:
            return None
        return self.textstr[self.pos]

    def peek(self, n: int = 1) -> str | None:
        p = self.pos + n
        if p >= self.len or p < 0:
            return None
        return self.buffer.text[p]

    def next(self) -> str | None:
        if self.atend():
            return None
        c = self.current
        self.move(1)
        return c

    def next_token(self) -> None:
        p = None
        while self.pos != p:
            p = self.pos
            self.eat_whitespace()
            if self.eat_eol_comments():
                self.eat_whitespace()
            self.eat_comments()

    def eat_whitespace(self) -> None:
        if self.buffer.whitespace_re:
            self._eat_regex(self.buffer.whitespace_re)

    def eat_comments(self) -> list[str]:
        return self._eat_regex_list(self.buffer.config.comments)

    def eat_eol_comments(self) -> list[str]:
        return self._eat_regex_list(self.buffer.config.eol_comments)

    def match(self, token: str) -> str | None:
        if not token:
            return None

        p = self.pos
        text = self.textstr[p : p + len(token)]
        if self.buffer.ignorecase:
            is_match = text.lower() == token.lower()
        else:
            is_match = text == token
        if not is_match:
            return None

        self.move(len(token))
        partial_match = (
            self.buffer.nameguard
            and self.is_name_char(self.current)
            and self.is_name(token)
        )
        if partial_match:
            self.goto(p)
            return None

        return token

    def matchre(self, pattern: str | re.Pattern) -> str | None:
        if not (match := self._scanre(pattern)):
            return None

        matched = match.group()
        token = str_from_match(match)
        self.move(len(matched))
        return token

    def matchname(self) -> str | None:
        return matchname(self)

    def matchint(self) -> int | None:
        return matchint(self)

    def matchuint(self) -> int | None:
        return matchuint(self)

    def matchfloat(self) -> float | None:
        return matchfloat(self)

    def matchbool(self) -> bool | None:
        return matchbool(self)

    def eat_spaces_no_newlines(self):
        p = None
        while self.pos != p:
            p = self.pos
            self.move(take_non_newline_whitespace_len(self.textstr, self.pos))
            if self.eat_eol_comments():
                self.move(take_non_newline_whitespace_len(self.textstr, self.pos))
            self.eat_comments()

    def matcheol(self) -> bool:
        mark = self.pos
        self.eat_spaces_no_newlines()
        eol_len = take_linebreak_len(self.textstr, self.pos)
        if eol_len is None:
            self.pos = mark
            return False
        self.move(eol_len)
        self.eat_spaces_no_newlines()
        return True

    def lineinfo(self, pos: int | None = None) -> LineInfo:
        if pos is None:
            pos = self.pos

        buf = self.buffer
        if not buf.linecache or not buf.lineindex:
            return LineInfo(
                source=buf.source,
                line=0,
                col=0,
                start=0,
                end=len(buf.text),
                text=buf.text,
            )

        # -2 to skip over sentinel
        pos = min(pos, len(buf.linecache) - 2)
        start, line, length = buf.linecache[pos]
        end = start + length
        col = pos - start

        text = buf.text[start:end]

        # only required to support includes
        n = min(len(buf.lineindex) - 1, line)
        source, line = buf.lineindex[n]

        return LineInfo(
            source=source,
            line=line,
            col=col,
            start=start,
            end=end,
            text=text,
        )

    def lookahead_pos(self) -> str:
        if self.atend():
            return ''
        info = self.lineinfo(self.pos)
        return '@%d:%d' % (info.line + 1, info.col + 1)

    def lookahead(self) -> str:
        if self.atend():
            return ''
        info = self.lineinfo(self.pos)
        text = info.text[info.col : info.col + 1 + 80]
        return self.buffer.split_block_lines(text)[0].rstrip()

    def lineat(self, pos: int | None = None) -> int:
        if pos is None:
            pos = self.pos
        return self.buffer.linecache[pos].lineno

    def poscol(self, pos: int | None = None) -> int:
        if pos is None:
            pos = self.pos
        start = self.buffer.linecache[pos].startpos
        return pos - start

    def get_line(self, n: int | None = None) -> str:
        if n is None:
            n = self.line
        return self.buffer.get_line(n)

    def get_lines(
        self,
        start: int | None = None,
        end: int | None = None,
    ) -> list[str]:
        return self.buffer.get_lines(start, end)

    def line_index(self, start: int = 0, end: int | None = None) -> list[LineIndexInfo]:
        return self.buffer.line_index(start, end)

    def is_name_char(self, c: str | None) -> bool:
        return c is not None and (c.isalnum() or c in self.buffer._namechar_set)

    def is_name(self, s: str) -> bool:
        if not s:
            return False
        goodstart = s[0].isalpha() or s[0] in self.buffer._namechar_set
        return goodstart and all(self.is_name_char(c) for c in s[1:])

    def _eat_regex(self, regex: str | re.Pattern | None) -> None:
        if not regex:
            return
        while self._matchre_fast(regex):
            pass

    def _eat_regex_list(self, regex: str | re.Pattern | None) -> list[str]:
        if not regex:
            return []

        r = cached_re_compile(regex)
        if r is None:
            return []

        result = []
        while x := self.matchre(r):
            result.append(x)
        return result

    def _matchre_fast(self, pattern: str | re.Pattern | None) -> bool:
        if not (match := self._scanre(pattern)):
            return False

        self.move(len(match.group()))
        return True

    def _scanre(self, pattern: str | re.Pattern | None) -> re.Match[Any] | None:
        cre = cached_re_compile(pattern)
        if cre is None:
            return None
        else:
            return cre.match(self.textstr, self.pos)

    def __len__(self) -> int:
        return self.len

    def __repr__(self) -> str:
        pos = self.pos
        return f'{typename(self)}({pos=})'


class Buffer(Text):
    def __init__(
        self,
        text: str,
        *,
        config: ParserConfig | None = None,
        **settings: Any,
    ):
        config = ParserConfig.new(config=config, **settings)
        assert isinstance(config, ParserConfig)
        self.config = config

        text = str(text)
        self.text = self.original_text = text

        self.whitespace_re = self.build_whitespace_re(config.whitespace)
        self.nameguard = (
            config.nameguard
            if config.nameguard is not None
            else bool(self.whitespace_re) or bool(config.namechars)
        )
        self._namechar_set = set(config.namechars or '')

        self.pos = 0
        self.len = 0
        self.lines: list[str] = []
        self.linecache: list[PosLine] = []
        self.lineindex: list[LineIndexInfo] = []

        self._preprocess()
        self._postprocess()

    def newcursor(self) -> Cursor:
        return BufferCursor(self, pos=self.pos)

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop('lock', None)
        return state

    @property
    def source(self) -> str:
        if not self.lineindex:
            return ''
        n = max(0, min(len(self.lineindex) - 1, self.line))
        filename, _line = self.lineindex[n]
        return filename

    @property
    def linecount(self) -> int:
        return linecount(self.text)

    @property
    def ignorecase(self) -> bool:
        return bool(self.config.ignorecase)

    @property
    def whitespace(self) -> str | UndefinedType | None:
        return self.config.whitespace

    @staticmethod
    def build_whitespace_re(whitespace: Any) -> re.Pattern | None:
        if whitespace is Undefined:
            return DEFAULT_WHITESPACE_RE
        if whitespace in {None, ''}:
            return None
        elif isinstance(whitespace, re.Pattern):
            return whitespace
        elif whitespace:
            return cached_re_compile(whitespace)
        else:
            return None

    def _preprocess(self, /, *_args: Any, **_kwargs: Any):
        lines, index = self._preprocess_block(self.source, self.text)
        self.lines = lines
        self.lineindex = index
        self.text = self.join_block_lines(lines)

    def _postprocess(self):
        cache, _count = PosLine.build_line_cache(self.lines, len(self.text))
        self.linecache = cache
        self.len = len(self.text)

    def _preprocess_block(
        self,
        name: str,
        block,
        /,
        **kwargs,
    ) -> tuple[list[str], list[LineIndexInfo]]:
        lines = self.split_block_lines(block)
        index = LineIndexInfo.block_index(name, len(lines))
        return self.process_block(name, lines, index, **kwargs)

    def split_block_lines(self, block: str) -> list[str]:
        return block.splitlines(True)

    def join_block_lines(self, lines: list[str]):
        return ''.join(lines)

    def process_block(
        self,
        _name: str,
        lines: list[str],
        index: list[LineIndexInfo],
        /,
        **_kwargs,
    ) -> tuple[list[str], list[LineIndexInfo]]:
        return lines, index

    def include(
        self,
        lines: list[str],
        index: list[LineIndexInfo],
        i: int,
        j: int,
        name: str,
        block: str,
        /,
        **kwargs,
    ) -> int:
        blines, bindex = self._preprocess_block(name, block, **kwargs)
        assert len(blines) == len(bindex)
        lines[i:j] = blines
        index[i:j] = bindex
        assert len(lines) == len(index)
        return j + len(blines) - 1

    def include_file(
        self,
        source: str,
        name: str,
        lines: list[str],
        index: list[LineIndexInfo],
        i: int,
        j: int,
    ) -> int:
        text, filename = self.get_include(source, name)
        return self.include(lines, index, i, j, filename, text)

    def get_include(self, source: str, filename: str) -> tuple[str, str]:
        source_path = Path(source).resolve()
        base = source_path.parent
        include = base / filename
        try:
            return include.read_text(), str(include)
        except OSError as e:
            raise ValueError(f'include not found: {include}') from e

    def replace_lines(self, i: int, j: int, name: str, block: str) -> tuple[int, str]:
        lines = self.split_block_lines(self.text)
        index = self.lineindex

        endline = self.include(lines, index, i, j, name, block)

        self.text = self.join_block_lines(lines)
        self.lineindex = index
        self._postprocess()

        newtext = self.join_block_lines(lines[j + 1 : endline + 2])
        return endline, newtext

    @property
    def line(self) -> int:
        return self.posline()

    @property
    def col(self) -> int:
        return self.poscol()

    def posline(self, pos: int | None = None) -> int:
        if not self.linecache:
            return 0
        if pos is None:
            pos = self.pos
        pos = max(0, min(pos, len(self.linecache) - 2))
        return self.linecache[pos].lineno

    def poscol(self, pos: int | None = None) -> int:
        if pos is None:
            pos = self.pos
        start = self.linecache[pos].startpos
        return pos - start

    def atend(self) -> bool:
        return self.pos >= self.len

    def ateol(self) -> bool:
        return self.atend() or self.current in {'\r', '\n', None}

    @property
    def current(self) -> str | None:
        if self.pos >= self.len:
            return None
        return self.text[self.pos]

    def at(self, p: int) -> str | None:
        if p >= self.len:
            return None
        return self.text[p]

    def peek(self, n: int = 1) -> str | None:
        return self.at(self.pos + n)

    def next(self) -> str | None:
        if self.atend():
            return None
        c = self.text[self.pos]
        self.move(1)
        return c

    def goto(self, pos: int):
        self.pos = max(0, min(len(self.text), pos))

    def move(self, n: int):
        self.goto(self.pos + n)

    def _eat_regex(self, regex: str | re.Pattern | None) -> bool:
        if not regex:
            return False
        seen = False
        while self._matchre_fast(regex):
            seen = True
        return seen

    def _eat_regex_list(self, regex: str | re.Pattern | None) -> list[str]:
        if not regex:
            return []

        r = cached_re_compile(regex)
        if r is None:
            return []

        def takewhile_repeat_regex():
            while x := self.matchre(r):
                yield x

        return list(takewhile_repeat_regex())

    def eat_whitespace(self) -> bool:
        if not self.whitespace_re:
            return False
        return self._eat_regex(self.whitespace_re)

    def eat_comments(self) -> bool:
        return self._eat_regex(self.config.comments)

    def eat_eol_comments(self) -> bool:
        return self._eat_regex(self.config.eol_comments)

    def next_token(self) -> None:
        p = None
        while self.pos != p:
            p = self.pos
            self.eat_whitespace()
            while self.eat_eol_comments():
                self.eat_whitespace()
            self.eat_comments()

    def skip_to(self, c: str) -> int:
        p = self.pos
        le = self.len
        while p < le and self.text[p] != c:
            p += 1
        self.goto(p)
        return self.pos

    def skip_past(self, c: str) -> int:
        self.skip_to(c)
        self.next()
        return self.pos

    def skip_to_eol(self) -> int:
        return self.skip_to('\n')

    def scan_space(self) -> bool:
        return bool(self.whitespace_re) and bool(self._scanre(self.whitespace_re))

    def is_space(self) -> bool:
        return self.scan_space()

    def is_name_char(self, c: str | None) -> bool:
        return c is not None and (c.isalnum() or c in self._namechar_set)

    def is_name(self, s: str) -> bool:
        if not s:
            return False
        goodstart = s[0].isalpha() or s[0] in self._namechar_set
        return goodstart and all(self.is_name_char(c) for c in s[1:])

    def match(self, token: str | None) -> str | None:
        if token is None:
            return None

        p = self.pos
        text = self.text[p : p + len(token)]
        if self.ignorecase:
            is_match = text.lower() == token.lower()
        else:
            is_match = text == token
        if not is_match:
            return None

        self.move(len(token))
        partial_match = (
            self.nameguard and self.is_name_char(self.current) and self.is_name(token)
        )
        if partial_match:
            self.goto(p)
            return None

        return token

    def _matchre_fast(self, pattern: str | re.Pattern | None) -> bool:
        if not (match := self._scanre(pattern)):
            return False

        self.move(len(match.group()))
        return True

    def matchre(self, pattern: str | re.Pattern) -> str | None:
        if not (match := self._scanre(pattern)):
            return None

        matched = match.group()
        token = str_from_match(match)
        self.move(len(matched))
        return token

    def _scanre(self, pattern: str | re.Pattern | None) -> re.Match[Any] | None:
        cre = cached_re_compile(pattern)
        if cre is None:
            return None
        else:
            return cre.match(self.text, self.pos)

    def indexed(self) -> bool:
        return bool(self.linecache) and bool(self.lineindex)

    def lineinfo(self, pos: int | None = None) -> LineInfo:
        if pos is None:
            pos = self.pos
        if not self.indexed():
            return LineInfo(
                source=self.source,
                line=0,
                col=0,
                start=0,
                end=len(self.text),
                text=self.text,
            )

        # -2 to skip over sentinel
        pos = min(pos, len(self.linecache) - 2)
        start, line, length = self.linecache[pos]
        end = start + length
        col = pos - start

        text = self.text[start:end]

        # only required to support includes
        n = min(len(self.lineindex) - 1, line)
        source, line = self.lineindex[n]

        return LineInfo(
            source=source,
            line=line,
            col=col,
            start=start,
            end=end,
            text=text,
        )

    def lookahead_pos(self) -> str:
        if self.atend():
            return ''
        info = self.lineinfo()
        return '@%d:%d' % (info.line + 1, info.col + 1)

    def lookahead(self) -> str:
        if self.atend():
            return ''
        info = self.lineinfo()
        text = info.text[info.col : info.col + 1 + 80]
        return self.split_block_lines(text)[0].rstrip()

    def get_line(self, n: int | None = None) -> str:
        if n is None:
            n = self.line
        return self.lines[n]

    def get_lines(self, start: int | None = None, end: int | None = None) -> list[str]:
        start = notnone(start, 0) or 0
        return self.lines[start : notnone(end, start + 1)]

    def line_index(self, start: int = 0, end: int | None = None) -> list[LineIndexInfo]:
        if end is None:
            return self.lineindex
        return self.lineindex[start : 1 + end]

    def __repr__(self) -> str:
        return f'{type(self).__name__}()'

    def __json__(self, _seen: set[int] | None = None) -> str | None:
        return None



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import inspect
from collections.abc import Callable
from contextlib import AbstractContextManager
from functools import cache
from typing import Any, Protocol, runtime_checkable

from ..exceptions import FailedParse
from ..input import Cursor, Text
from .ast import AST
from .infos import RuleInfo
from .state import ParseState, ParseStateStack


type Func = Callable[[Ctx], Any]


@cache
def is_func(func: Callable) -> bool:
    declared = inspect.signature(func).parameters
    if len(declared) == 1:
        p, *_ = declared.values()
        return p.annotation == 'Ctx'
    # if len(declared) == 2:
    #     p1, p2, *_ = declared.values()
    #     return p1.name == 'self' and p2.annotation == 'Ctx'
    return False


@runtime_checkable
class CanParse(Protocol):
    def parse(
        self,
        text: str | Text,
        /,
        *,
        start: str | None = None,
        config: Any = None,
        asmodel: bool = False,
        **settings: Any,
    ) -> Any: ...


@runtime_checkable
class Ctx(CanParse, Protocol):
    states: ParseStateStack

    @property
    def ast(self) -> AST: ...

    @property
    def cst(self) -> Any: ...

    @property
    def last_node(self) -> Any: ...

    @property
    def state(self) -> ParseState: ...
    def statescope(self, *, merge: bool = True) -> Any: ...

    @property
    def cursor(self) -> Cursor: ...
    @property
    def callstack(self) -> list[RuleInfo]: ...
    def heartbeat(self) -> bool: ...
    def newexcept(
        self,
        msg: Any,
        excls: type[FailedParse] = FailedParse,
    ) -> FailedParse: ...

    def call(self, ri: RuleInfo) -> Any: ...
    def find_rule(self, name: str) -> Func: ...

    def alert(self, message: str, level: int) -> None: ...
    def choice(self) -> Any: ...
    def constant(self, literal: Any) -> Any: ...
    def cut(self) -> None: ...
    def define(self, keys: list[str], addkeys: list[str] | None = None) -> None: ...
    def dot(self) -> Any: ...
    def eofcheck(self) -> None: ...
    def eolcheck(self) -> None: ...
    def empty(self) -> list: ...
    def fail(self): ...
    def gatheropt(self) -> Any: ...
    def gatherplus(self) -> Any: ...
    def group(self) -> Any: ...
    def groupexp(self, exp: Func) -> Any: ...
    def skipgroup(self) -> Any: ...
    def if_(self) -> Any: ...
    def ifnot_(self) -> Any: ...
    def joinopt(self) -> Any: ...
    def joinplus(self) -> Any: ...
    def joinleft(self) -> Any: ...
    def joinright(self) -> Any: ...
    def loopopt(self) -> Any: ...
    def loopplus(self) -> Any: ...
    def nameset(self, name: str) -> Any: ...
    def nameadd(self, name: str) -> Any: ...
    def result(self) -> Any: ...
    def resultadd(self) -> Any: ...
    def option(self) -> Any: ...
    def optional(self) -> Any: ...
    def token(self, token: str) -> str: ...
    def pattern(self, pattern: str) -> Any: ...
    def matchname(self) -> str | None: ...
    def matchint(self) -> int | None: ...
    def matchuint(self) -> int | None: ...
    def matchfloat(self) -> float | None: ...
    def matchbool(self) -> float | None: ...
    def skipto(self) -> Any: ...
    def skip_to(self, exp: Func) -> Any: ...
    def void(self) -> Any: ...

    def closure(
        self,
        exp: Func,
        sep: Func | None = None,
        omitsep: bool = False,
    ) -> Any: ...
    def positive_closure(
        self,
        exp: Func,
        sep: Func | None = None,
        omitsep: bool = False,
    ) -> Any: ...
    def join(self, exp: Func, sep: Func) -> Any: ...
    def positive_join(self, exp: Func, sep: Func) -> Any: ...
    def left_join(self, exp: Func, sep: Func) -> Any: ...
    def right_join(self, exp: Func, sep: Func) -> Any: ...
    def gather(self, exp: Func, sep: Func) -> Any: ...
    def positive_gather(self, exp: Func, sep: Func) -> Any: ...
    def isolate(self, exp: Func) -> Any: ...
    def expcall(self, exp: Func) -> Any: ...

    def bound(
        self,
        text: str | Text,
        /,
        *,
        start: str | None = None,
        config: Any = None,
        asmodel: bool = False,
        **settings: Any,
    ) -> AbstractContextManager[Ctx]: ...



# ruff: noqa: F811
# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import functools
from collections.abc import Callable
from typing import Any, cast

from ..ctx import Ctx
from ..infos import RuleInfo


def __rule_wrapper(func: Callable, *args: Any, **kwargs: Any) -> Callable:
    ribase = RuleInfo.new(None, func, args, kwargs)

    @functools.wraps(func)
    def wrapper(instance: Any, ctx: Ctx) -> Any:
        ri = RuleInfo.bind(ribase, instance)
        return ctx.call(ri)

    return wrapper


def rule(*args: Any, **kwargs: Any) -> Callable:
    if len(args) == 1 and callable(args[0]) and not kwargs:
        func: Callable = cast(Callable, args[0])
        return __rule_wrapper(func)

    def decorator(func: Callable) -> Callable[[Any, Ctx], Any]:
        return __rule_wrapper(func, *args, *kwargs)

    return decorator



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import inspect
from dis import Positions
from pathlib import Path


def execute_logic(a, b) -> float:
    # Multiple lines to fill the context buffer
    val_a = a + 10
    val_b = b + 20

    # The error occurs here
    calc = (val_a * 2) / (val_b - 20)

    return calc


def capture_error() -> Exception:
    try:
        execute_logic(5, 0)
        return Exception()
    except ZeroDivisionError as e:
        return e


def report_with_full_context(exc: Exception, size=4) -> None:
    tb = exc.__traceback__
    if tb is None:
        return
    # Fetch 'size' lines above and below
    frames = inspect.getinnerframes(tb, size)

    print(f"--- Traceback (Context Size: {size}) ---")
    for info in frames:
        p = Path(info.filename)
        pos = info.positions
        assert isinstance(pos, Positions)

        print(f"\nFile: {p.name}, Line {info.lineno}, in {info.function}")

        if info.code_context:
            for i, line in enumerate(info.code_context):
                # Use a marker for the current line in the gutter
                gutter = ">> " if i == info.index else "   "
                print(f"{gutter}{line.rstrip()}")

                # If this is the failing line, insert the column markers below it
                if i == info.index and pos.col_offset is not None:
                    # 3 spaces for the gutter + col_offset spaces
                    padding = "   " + (" " * pos.col_offset)

                    # Calculate caret length
                    low = pos.col_offset
                    high = pos.end_col_offset or low + 1
                    marker = "^" * max(1, high - low)

                    print(f"{padding}{marker}")
        else:
            print("   (Source code not available)")


if __name__ == "__main__":
    error = capture_error()
    report_with_full_context(error)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""TatSu CLI package, mirroring ogopego's three-subcommand structure."""

from __future__ import annotations

from .cli import cling_main


__all__ = [
    "cling_main",
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import dataclasses as dc
import inspect
import warnings
from collections.abc import Callable, Iterable, Mapping
from functools import cache
from typing import Any, Self, overload

from ..contexts.infos import ParseInfo
from ..util import rowselect, typename
from ..util.asjson import AsJSONMixin, asjson, asjsons
from ..util.fromjson import JSONBase
from ..util.indent import fold


__all__ = ['BaseNode', 'NodeDataclassParams', 'nodedataclass']

NodeDataclassParams = dict(
    {},
    eq=False,
    repr=False,
    match_args=False,
    unsafe_hash=False,
    kw_only=True,
)


@overload
def nodedataclass[T: type](cls: T) -> T: ...


@overload
def nodedataclass[T: type](**params: Any) -> Callable[[T], T]: ...


def nodedataclass[T: type](cls: T | None = None, **params: Any) -> T | Callable[[T], T]:
    # by Gemini (2026-02-07)
    # by [apalala@gmail.com](https://github.com/apalala)

    def decorator(target: T) -> T:
        allparams = {**NodeDataclassParams, **params}
        return dc.dataclass(**allparams)(target)  # type: ignore

    # If cls is passed, it was used as @tatsudataclass with no arguments
    if cls is not None and not params:
        return decorator(cls)

    return decorator


@nodedataclass
class BaseNode(JSONBase, AsJSONMixin):
    ast: Any = dc.field(kw_only=False, default=None)
    ctx: Any = None
    parseinfo: ParseInfo | None = None

    def __init__(self, ast: Any = None, **attributes: Any):
        # NOTE:
        #  A @datclass subclass may not call this,
        #  but __post_init__() should still be honored
        super().__init__()

        self.ast = ast
        self.ctx = None
        self.parseinfo = None
        self.__set_attributes(**attributes)
        self.__post_init__()

    def __post_init__(self):
        ast = self.ast
        if not isinstance(ast, dict):
            return

        # note:
        #   Node objects are created by a model builer when invoked by he parser,
        #   which passes only the ast recovered when the object was created.
        #   `
        #       point::Point = ... left:... right:... ;
        #   `
        #   Here the key,value pairs in the AST are injected into the corresponding
        #   attributes declared by the Node subclass. Synthetic classes
        #   override this to create the attributes.
        keys = self._in_field_order(ast)
        for name in keys:
            if not hasattr(self, name) or inspect.ismethod(getattr(self, name)):
                continue
            setattr(self, name, ast[name])

    def __set_attributes(self, **attrs) -> None:
        if not isinstance(attrs, dict):
            return

        for name, value in attrs.items():
            if not hasattr(type(self), name):
                raise ValueError(f'Unknown argument {name}={value!r}')
            if inspect.isroutine(method := getattr(type(self), name)):
                raise TypeError(f'Overriding method {name}={method!r}')

            if (prev := getattr(self, name, None)) and inspect.ismethod(prev):
                warnings.warn(
                    f'`{name}` in keyword arguments will override'
                    f' `{typename(self)}.{name}`: {prev!r}',
                    stacklevel=2,
                )
            setattr(self, name, value)

    def clone(self, **attrs: Any) -> Self:
        return dc.replace(self, **attrs)  # type: ignore # pyright: ignore[reportArgumentType]

    def asjson(self) -> Any:
        return asjson(self)

    def asjsons(self) -> str:
        return asjsons(self.asjson())

    def dump(self) -> Any:
        return self.asjson()

    def dumps(self) -> str:
        return self.asjsons()

    @staticmethod
    @cache
    def _basenode_keys() -> frozenset[str]:
        # Gemini (2026-02-14)
        return frozenset(vars(BaseNode).keys())

    def __pub__(self, sunderok: bool = False) -> dict[str, Any]:
        pub = super().__pub__(sunderok=sunderok)
        if sunderok:
            return pub  # we're being called from __getstate__ not __repr__

        # Gemini (2026-02-14)
        wanted = pub.keys() - self._basenode_keys()
        if wanted or self.ast is None:
            pass
        elif not isinstance(self.ast, dict):
            wanted = {'ast'}  # self.ast may be all this object has

        pub = rowselect(wanted, pub)
        sortedkeys = self._in_field_order(pub.keys())
        return {name: pub[name] for name in sortedkeys if name in pub}

    def _in_field_order(self, keys: Iterable[str] | None) -> list[str]:
        if keys is None:
            keys = dir(self)
        fieldindex = {f.name: i for i, f in enumerate(dc.fields(self))}  # type: ignore

        def fieldorder(n) -> int:
            return fieldindex.get(n, len(fieldindex))

        return sorted(keys, key=fieldorder)

    def __repr__(self) -> str:
        values = self.__pub__()

        attr_repr_list = []
        for name, value in values.items():
            if value is None:
                continue

            if len(values) == 1:
                prefix = ''
            else:
                prefix = f"{name}="
            line = fold(
                prefix=prefix,
                value=value,
                amount=2,
                addlevels=4,
            )
            attr_repr_list += [line]

        return fold(
            prefix=f'{typename(self)}',
            value=tuple(attr_repr_list),
            reprs=False,
            amount=2,
            addlevels=2,
        )

    def __str__(self) -> str:
        return super().__repr__()

    def __eq__(self, other) -> bool:
        # NOTE: No use case for structural equality
        return other is self

    def __hash__(self) -> int:
        # NOTE: No use case for structural equality
        return hash(id(self))

    def __getstate__(self) -> Any:
        return self.__pub__(sunderok=True)

    def __setstate__(self, state: dict[str, Any]) -> None:
        for name, value in state.items():
            try:
                setattr(self, name, value)
            except AttributeError:
                if not hasattr(self, name):
                    raise
                # a property without a setter ?
                pass

    @classmethod
    def __from_json__(cls, data: Mapping[str, Any]) -> Self:
        return super().__from_json__(data)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import inspect
from functools import cache
from typing import Any

from .config import ParserConfig
from .contexts import (
    Func,
    ParseContext,
    isname,
    leftrec,
    name,
    nomemo,
    tatsumasu,
    tatsumasu as rule,  # note: this supports v5.16-v5.17 parsers
)
from .exceptions import FailedRef
from .util import generic_main, safe_name


__all__ = [
    'Parser',
    'NGParser',
    'generic_main',
    'isname',
    'name',
    'leftrec',
    'nomemo',
    'rule',
    'tatsumasu',
]


@cache
def find_rule(source: Any, name: str) -> Func | None:
    for rulename in {name, name.strip('_'), f'_{name}_', f'_{name}'}:
        action = getattr(source, safe_name(rulename), None)
        if callable(action):
            return action
    return None


class Parser(ParseContext):
    def __init__(
        self,
        rulesource: Any = None,
        /,
        *,
        config: ParserConfig | None = None,
        **settings: Any,
    ) -> None:
        self.rulesource = rulesource

        config = ParserConfig.new(config, **settings)
        srcconfig = ParserConfig.new(getattr(rulesource, '_config', None))
        config = srcconfig.override_config(config)

        super().__init__(config=config)

    def find_rule(self, name: str) -> Func:
        if self.rulesource:
            action = find_rule(self.rulesource, name)
        else:
            action = find_rule(self, name)
        if not action:
            raise self.newexcept(f'{name}', excls=FailedRef)
        return action

    def rule_list(self) -> list[str]:
        source = self.rulesource or type(self)

        def isdunder(name: str) -> bool:
            return name.startswith('__') and name.endswith('__')

        methods = inspect.getmembers(source, predicate=inspect.ismethod)
        return [m[0] for m in methods if not isdunder(m[0])]


# NOTE: backwards compatibility
NGParser = Parser



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import sys
import threading
from collections.abc import Generator
from contextlib import contextmanager


@contextmanager
def background() -> Generator[None, None, None]:
    """Executes the inner code block on a separate thread using shared memory.
    Guarantees thread cleanup and exception propagation."""

    thread_exception = []
    block_generator = yield

    def run_target():
        try:
            next(block_generator)
        except StopIteration:
            pass
        except Exception:
            thread_exception.append(sys.exc_info())

    worker = threading.Thread(target=run_target, daemon=True)
    worker.start()

    worker.join()
    if thread_exception:
        _exc_type, exc_val, exc_tb = thread_exception[0]
        raise exc_val.with_traceback(exc_tb)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Iterator
from typing import Any

from ..ctx import Ctx, Func
from .exp import ExpContext


class LoopContext(ExpContext):
    def __init__(self, ctx: Ctx, plus: bool = False):
        super().__init__(ctx)
        self.plus = plus

    def _inner_iter(self, ctx: Ctx, func: Func) -> Iterator[Any]:
        yield ctx.isolate(func)

    def parse(self, ctx: Ctx) -> Any:
        if self.plus:
            return ctx.positive_closure(self.func)
        else:
            return ctx.closure(self.func)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Any

from .objectmodel.builder import ModelBuilderSemantics
from .util import simplify_list
from .util.typetools import TypeContainer


__all__ = ['ASTSemantics', 'ModelBuilderSemantics', 'TypeContainer']


class ASTSemantics:
    def group(self, ast: Any, *args) -> Any:
        _ = args
        return simplify_list(ast)

    def element(self, ast: Any, *args) -> Any:
        _ = args
        return simplify_list(ast)

    def sequence(self, ast: Any, *args) -> Any:
        _ = args
        return simplify_list(ast)

    def choice(self, ast: Any, *args) -> Any:
        _ = args
        if len(ast) == 1:
            return simplify_list(ast[0])
        return ast



# Copyright © 2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re
from enum import Enum
from typing import NamedTuple, assert_never

from .style import DEFAULT_COLOR, Color, Style


class StyleZ(str):
    __slots__ = ("_color", "_styles")

    def __new__(cls, *args, **_kwargs):
        assert args and isinstance(args[0], list)
        styles = args[0]
        assert not styles or isinstance(styles[0], Style)
        value = "".join(s.value for s in styles)
        new = super().__new__(cls, value)
        new._styles = styles
        return new

    def __init__(self, styles: list[Style]):
        self._styles = styles

    @property
    def value(self) -> str:
        return super().__str__()

    @property
    def style(self) -> list[Style]:
        return self._styles

    def __str__(self) -> str:
        return "".join(str(s) for s in self._styles)

    def __repr__(self) -> str:
        return "".join(repr(s) for s in self._styles)


class TokenType(Enum):
    ESC = "ESC"
    RET = "RET"
    CAL = "CAL"
    TXT = "TXT"


class Token(NamedTuple):
    type: TokenType
    value: str


def apply_style_stack(
    stack: list[str],
    text: str,
    color: Color = DEFAULT_COLOR,
) -> Style:
    from .xstyle import XStyle

    s = XStyle(text, color=color)

    for name in stack:
        if style := s.get_style_method(name):
            s = style()
    assert s is not None
    return s


def markup(*texts: str, color: Color = DEFAULT_COLOR) -> StyleZ:
    stack: list[str] = []

    out: list[Style] = []
    part: str = ""
    for text in texts:
        text = str(text)  # noqa: PLW2901
        for token in tokenize(text):
            if part and token.type != TokenType.TXT:
                out += [apply_style_stack(stack, part, color=color)]
                part = ""
            match token.type:
                case TokenType.CAL:
                    for style in token.value.split(" "):
                        stack += [style]
                case TokenType.RET:
                    if not token.value:
                        stack = stack[:-1]
                    elif token.value in {"all", "*"}:
                        stack = []
                    elif stack and stack[-1] == token.value:
                        stack.pop()
                case TokenType.ESC:
                    part += "["
                case TokenType.TXT:
                    part += token.value
                case _:
                    assert_never(token.type)

    if part:
        out += [apply_style_stack(stack, part, color=color)]

    return StyleZ(out)


def tokenize(text: str) -> list[Token]:
    """Tokenize a Rich markup string into unbracketed tag data and text."""
    pattern = r"""(?x)
        (?P<ESC>\[\[)
        |
        \[/(?P<RET>[^\]]*)\]
        |
        \[(?P<CAL>[^\]]+)\]
        |
        (?P<TXT>[^\[]+)
    """

    tokens = []
    for match in re.finditer(pattern, text):
        kind = match.lastgroup
        assert kind is not None
        val = match.group(kind)

        match kind:
            case "CAL":
                tokens.append(Token(TokenType.CAL, val))
            case "RET":
                tokens.append(Token(TokenType.RET, val))
            case "TXT":
                tokens.append(Token(TokenType.TXT, val))
            case "ESC":
                tokens.append(Token(TokenType.TXT, "["))
            case _ as never:
                raise AssertionError(f"Unexpected kind: {never}")

    return tokens


def tokenize_to_values(text: str) -> list[str]:
    tokens = tokenize(text)
    return [t.value for t in tokens]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import asyncio
import atexit
import contextlib
import json
import os
import sys
from collections.abc import AsyncGenerator, Generator, Iterator
from pathlib import Path
from typing import IO, Any

from tatsu.util.fromjson import JSONBase

from ..util import alpha_timestamp
from .packet import (
    BadPacketError,
    Packet,
    PacketHashError,
    PacketLike,
    pack,
    unpack,
)


PACKETZ_DIR = Path(f"./.{__name__.split('.')[-2]}")

_defer_deinit_queues: set[PacketzQueue] = set()


def new_file_path() -> Path:
    return PACKETZ_DIR / f"{alpha_timestamp()}.pktz.jsonl"


class PacketzQueue(JSONBase):
    """File-backed packet queue.

    Instantiate with a path, or let it generate a timestamped file
    under ``.packetz/``.
    """

    def __init__(self, /, path: Path | str | None = None, *, keep: bool | None = None):
        if path is None:
            path = new_file_path()
            if not self._should_keep(keep):
                atexit.register(_cleanup_queue, self)
        self.path = Path(path)

        PACKETZ_DIR.mkdir(parents=True, exist_ok=True)
        self.path.parent.mkdir(parents=True, exist_ok=True)

        if not self.path.exists():
            self.path.touch(exist_ok=True)
        self._told = 0
        self._seen: set[str] = set()

    def _should_keep(self, keep: bool | None) -> bool:
        if keep is None:
            keep = bool(k := os.environ.get("PACKETZ_KEEP", "").lower()) and (
                k
                not in {
                    "0",
                    "false",
                    "no",
                }
            )
        return keep

    def _queue_healthy(self, q: IO[str] | None) -> bool:
        if q is sys.stdin or q is sys.stdout:
            return False
        if q is None or q.closed:
            return False
        return os.fstat(q.fileno()).st_nlink > 0

    def _ensure_open(self) -> IO[str]:
        q = self.path.open(
            "rt",
            encoding="utf-8",
            buffering=1024 * 256,
        )
        assert self._queue_healthy(q)
        return q

    def _get_queue(self) -> IO[str]:
        return self._ensure_open()

    def tell(self, q: IO[str], check: bool = True) -> None:
        t = q.tell()
        assert (check and t > self._told) or (not check and t >= self._told)
        self._told = t

    @contextlib.contextmanager
    def reader(self) -> Generator[IO[str], None, None]:
        with self._get_queue() as q:
            try:
                q.seek(self._told)
                yield q
            finally:
                self.tell(q, False)

    @contextlib.contextmanager
    def writer(self) -> Generator[IO[str], None, None]:
        with self.path.open("at", encoding="utf-8", buffering=1) as queue:
            yield queue

    def send(self, *, to: str | None = None, data: Any = None) -> PacketLike:
        packet = Packet(to=to, data=data)
        serial = pack(packet)
        with self.writer() as queue:
            queue.write(serial + "\n")
        return packet

    def receive(self) -> Iterator[PacketLike]:
        with self._ensure_open() as q:
            q.seek(self._told)

            while line := q.readline():
                # If the line doesn't end with a newline, it's a partial write.
                # Stop here and leave self._told right at the start of this line.
                if not line.endswith("\n"):
                    break

                self._told = max(q.tell(), self._told)

                try:
                    packet = unpack(line)
                except (
                    BadPacketError,
                    json.JSONDecodeError,
                    TypeError,
                    PacketHashError,
                    ValueError,
                ):
                    continue  # Skip corrupt rows safely

                if packet.id not in self._seen:
                    self._seen.add(packet.id)
                    yield packet

    def receive_0(self) -> Iterator[PacketLike]:
        with self.reader() as queue:
            lines = queue.readlines()
        for serial in lines:
            try:
                packet = unpack(serial)
            except (BadPacketError, json.JSONDecodeError, TypeError):
                continue  # NOTE Incomplete? Wait and retry
            except (PacketHashError, ValueError):
                continue  # NOTE Skip and continue!
            if packet.id in self._seen:
                continue
            self._seen.add(packet.id)
            yield packet

    async def receive_async(self) -> AsyncGenerator[PacketLike, None]:
        try:
            while True:
                for msg in self.receive():
                    yield msg
                await asyncio.sleep(0.01)
        except asyncio.CancelledError:  # noqa: TRY203
            raise


def _cleanup_queue(q: PacketzQueue):
    path = q.path
    with contextlib.suppress(OSError):
        path.unlink(missing_ok=True)
    with contextlib.suppress(OSError):
        PACKETZ_DIR.rmdir()



# Copyright (c) 2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import argparse
import ast
import sys
from itertools import starmap
from pathlib import Path
from typing import NamedTuple

from ..treez import Tree
from ..util.moduletools import pathtomodulename
from ..ztyle import Color, Style


color = Color.default()


class Dependency(NamedTuple):
    name: str
    is_relative: bool


class ModuleImports(NamedTuple):
    name: str
    path: Path
    imports: tuple[Dependency, ...]


class DependencyNodeKind(NamedTuple):
    symbol: str | None
    style: Style
    prefix: str = ""
    suffix: str = ""
    sort_rank: int = 0

    @classmethod
    def sibling(cls) -> DependencyNodeKind:
        return cls("○", color.style().cyan(), "", "", 1)

    @classmethod
    def internal(cls) -> DependencyNodeKind:
        return cls("◉", color.style().cyan(), "", "", 2)

    @classmethod
    def external(cls) -> DependencyNodeKind:
        return cls(None, color.style().white(), "⟨", "⟩", 3)


def programfiles() -> list[Path]:
    """
    Return all Python files that belong to the current program's
    top-level package, if available.
    """
    if not __package__:
        return []

    import importlib

    rootpkgname = __package__.split(".")[0]
    rootpkg = importlib.import_module(rootpkgname)
    rootfile = getattr(rootpkg, "__file__", None)
    if not rootfile:
        return []

    rootpath = Path(rootfile).parent.relative_to(Path.cwd())
    return [p for p in rootpath.rglob("*.py") if p.is_file()]


def moduledeps(name: str, path: Path) -> ModuleImports:
    """
    Parses a Python file to find all its module-level imports, resolving
    relative imports to their fully qualified names.
    """
    source = path.read_text()
    module = ast.parse(source, filename=name)
    assert isinstance(module, ast.Module), module

    found_imports: set[Dependency] = set()

    package_name_parts = name.split('.')
    if path.name != '__init__.py':
        package_name_parts = package_name_parts[:-1]

    for node in module.body:
        if not isinstance(node, ast.Import | ast.ImportFrom):
            continue

        if isinstance(node, ast.Import):
            found_imports.update(
                Dependency(alias.name, is_relative=False) for alias in node.names
            )
            continue

        if node.module == "__future__":
            continue

        level = node.level
        if level == 0:
            if node.module:  # It's an absolute import
                found_imports.add(Dependency(str(node.module), is_relative=False))
            continue

        # It's a relative import
        dot_count = level - 1
        if dot_count > len(package_name_parts):
            continue

        relative_base_parts = package_name_parts[: len(package_name_parts) - dot_count]
        base = ".".join(relative_base_parts)

        if node.module:
            full_name = f"{base}.{node.module}" if base else node.module
            found_imports.add(Dependency(full_name, is_relative=True))
            continue

        for alias in node.names:
            full_name = f"{base}.{alias.name}" if base else alias.name
            found_imports.add(Dependency(full_name, is_relative=True))

    return ModuleImports(
        name=name,
        path=path,
        imports=tuple(sorted(found_imports, key=lambda x: x.name)),
    )


def findeps(paths: list[Path]) -> list[ModuleImports]:
    module_data = sorted((pathtomodulename(path), path) for path in paths)
    return list(starmap(moduledeps, module_data))


def add_dependency_node(
    parent: Tree,
    qualified_name: str,
    is_internal: bool,
    importer_name: str,
    internal_roots: set[str],
):
    """
    Adds a dependency as an intelligently-named leaf node to the tree.
    """
    if not is_internal:
        kind = DependencyNodeKind.external()
        label = f"{kind.prefix}{qualified_name}{kind.suffix}"
    else:
        kind = DependencyNodeKind.internal()
        display_name = qualified_name

        importer_parts = importer_name.split('.')
        importee_parts = qualified_name.split('.')

        common_len = 0
        min_len = min(len(importer_parts), len(importee_parts))
        while (
            common_len < min_len
            and importer_parts[common_len] == importee_parts[common_len]
        ):
            common_len += 1

        importer_pkg_depth = len(importer_parts) - 1

        if (
            common_len == importer_pkg_depth
            and len(importee_parts) == importer_pkg_depth + 1
        ):
            display_name = importee_parts[-1]
            kind = DependencyNodeKind.sibling()
        elif common_len == importer_pkg_depth - 1:
            suffix = ".".join(importee_parts[common_len:])
            display_name = f"..{suffix}"
        elif qualified_name.startswith(importer_name + '.'):
            display_name = qualified_name[len(importer_name) + 1 :]
        else:
            root_part = qualified_name.split('.', 1)[0]
            if root_part in internal_roots:
                display_name = qualified_name[len(root_part) + 1 :]
                display_name = display_name.removeprefix(".")

        if kind.sort_rank == 1:
            display_name = f".{display_name}"
        label = f"{kind.symbol} {display_name}"

    label = str(kind.style(label))
    parent.add(label)


def render(results: list[ModuleImports]) -> Tree:
    """
    Renders the complete dependency tree from the analysis results.
    """
    root = Tree("Module Dependency Analysis")
    module_nodes: dict[str, Tree] = {}
    analyzed_module_names = {m.name for m in results}
    internal_roots = {m.name.split(".")[0] for m in results}

    for module_name in sorted(analyzed_module_names):
        current = root
        parts = module_name.split('.')
        for i, part in enumerate(parts):
            path_so_far = ".".join(parts[: i + 1])
            if path_so_far in module_nodes:
                current = module_nodes[path_so_far]
            else:
                new_node = current.add(str(color.style().cyan()(part)))
                module_nodes[path_so_far] = new_node
                current = new_node

    for module in results:
        module_node = module_nodes[module.name]
        for imp in module.imports:
            imp_parent = ".".join(imp.name.split('.')[:-1])
            if imp.name in analyzed_module_names and imp_parent == module.name:
                continue

            imp_root = imp.name.split(".")[0]
            is_internal_imp = imp.is_relative or (imp_root in internal_roots)
            add_dependency_node(
                module_node,
                imp.name,
                is_internal_imp,
                importer_name=module.name,
                internal_roots=internal_roots,
            )

    def sort_tree(tree: Tree):
        def sort_key(n: Tree):
            label = str(n.label)
            symbol_map = {
                DependencyNodeKind.sibling().symbol: DependencyNodeKind.sibling().sort_rank,
                DependencyNodeKind.internal().symbol: DependencyNodeKind.internal().sort_rank,
                DependencyNodeKind.external().prefix: DependencyNodeKind.external().sort_rank,
                None: 0,
            }
            key = label[0] if label else None
            return symbol_map.get(key, 0), label

        tree.children.sort(key=sort_key)
        for child in tree.children:
            sort_tree(child)

    sort_tree(root)

    if len(root.children) == 1:
        return root.children[0]

    return root


def add_ideps_cmd(sub) -> argparse.ArgumentParser:
    parser = sub.add_parser(
        "ideps",
        help="Analyze dependencies of Python modules",
    )
    add_argparse_options(parser)
    return parser


def add_argparse_options(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "files",
        nargs="*",
        help="Python files or glob patterns to analyze (default: auto-detect program files)",
    )


def ideps_cmd(parser: argparse.ArgumentParser) -> int:
    add_argparse_options(parser)
    args = parser.parse_args()

    color_opt = args.color
    if color_opt == "always":
        color.enable(True)
    elif color_opt == "never":
        color.enable(False)

    paths: list[Path] = []
    if not args.files and __package__:
        paths = programfiles()
    elif not args.files:
        prog = f"python {Path(__file__).name}"
        parser.print_usage()
        print(f"{prog}: error: no files specified and no program package detected")
        raise SystemExit(1)

    for arg in args.files:
        if any(c in arg for c in "*?[]"):
            expanded = Path().glob(arg)
            paths.extend(Path(p) for p in expanded if Path(p).is_file())
        elif (path := Path(arg)) and path.is_file():
            paths.append(path)

    paths = sorted(set(paths))

    if not paths:
        print("No Python files found matching the given paths.", file=sys.stderr)
        raise SystemExit(1)

    results = findeps(paths)
    tree = render(results)
    print(tree)
    return 0


def main() -> int:
    parser = argparse.ArgumentParser(
        description="Analyze and display Python module dependencies"
    )
    parser.add_argument(
        "-c",
        "--color",
        choices=["auto", "always", "never"],
        default="auto",
        help="Control colorized output (default: auto)",
    )
    return ideps_cmd(parser)


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import sys
import time
from collections.abc import Generator, Iterable
from pathlib import Path
from typing import Any

from ..log import iso_logpath, logctx, startscript
from ..packetz import PacketLike
from ..util import debugging, identity
from .parproc import Progress, parproc
from .payload import VisualPayload
from .result import Result
from .summary import show_result, show_summary
from .task import Func, VisualFunc


def parproc_visual(
    func: VisualFunc,
    payloads_in: Iterable[VisualPayload | str],
    /,
    progress_in: Progress | None = None,
    *args: Any,
    eprint: Func = debugging.eprint,
    pickable: Func = identity,
    parallel: bool = True,
    reraise: bool = False,
    summary: bool = True,
    verbose: bool = True,
    max_workers: int | None = None,
    usecolor: bool = True,
    **kwargs: Any,
) -> Generator[PacketLike, None, None]:
    from ..barz import BarRow, Col, Multi
    from ..util.tty import hide_cursor, show_cursor
    from ..ztyle import Color

    eprint_in = eprint

    # NOTE resolve iterator now because we know that processing will do it anyway
    payloads_in = list(payloads_in)
    total = len(payloads_in)

    style = Color(usecolor).style()
    f = style.green()

    b = style.black().bold()

    multi = None
    if progress_in is not None:
        progress: Progress = progress_in  # type: ignore # pyright: ignore[reportAssignmentType]
    else:
        progress = BarRow(
            total=total,
            fill="---",
            style=[f, f, b],
            cols=["   ", Col.label, Col.bar],
            width=47,
        )
        multi = Multi([], out=sys.stderr)
        eprint = multi.print

        multi.add_row(progress)
        progress.start()
        multi.start()

        print(hide_cursor(), end='\r')

    # HACK backwards compatibility
    is_legacy = all(isinstance(p, str) for p in payloads_in)
    payloads: list[VisualPayload] = (  # type: ignore # pyright: ignore[reportAssignmentType]
        payloads_in  # type: ignore
        if not is_legacy
        else [VisualPayload(path=Path(str(p)), payload=None) for p in payloads_in]
    )
    filenames = [str(p.path) for p in payloads]

    logpath = None
    if len(filenames) > 1:
        prefix = startscript().replace('.', '_')
        logpath = iso_logpath(prefix=prefix)

    start_time = time.time()
    packets: Iterable[PacketLike] = parproc(
        func,
        payloads,
        *args,
        pickable=pickable,
        parallel=parallel,
        reraise=reraise,
        max_workers=max_workers,
        **kwargs,
    )

    def process_packets(results: Iterable[PacketLike]) -> Generator[PacketLike]:
        count = 0
        for packet in results:
            if packet is None:
                continue
            if not isinstance(packet, Result):
                yield packet
                continue

            result: Result = packet
            count += 1
            path = result.payload.path
            progress.update(count, total, label=f"{path.name:44}")
            if verbose:
                show_result(eprint, result, width=44)

            if result.exception:
                if logpath:
                    with logctx(logpath) as log:
                        print('ERROR:', result.payload, file=log)
                        print(result.exception, file=log)
                else:
                    eprint(f'ERROR: {result.payload}')
                    eprint(result.exception)
                if reraise:
                    raise result.exception

            if is_legacy:
                result.payload = result.payload.path
            yield result

    packets = process_packets(packets)
    if is_legacy:
        packets = list(packets)

    if summary or is_legacy:
        results: list[Result] = [p for p in packets if isinstance(p, Result)]
        packets = show_summary(
            start_time,
            results,
            eprint=eprint_in,
            verbose=verbose,
        )

    if multi is not None:
        multi.print(show_cursor())
        multi.print()
        multi.stop()

    yield from packets



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import sys
import time
from collections.abc import Callable, Iterable, Mapping
from threading import Event
from typing import Any, NamedTuple

from ..util import memory_use
from .payload import VisualPayload
from .result import Result


type Func = Callable[[Any], Any]
type TaskFunc[*P] = Callable[[Task, *P], Any]
type TransfFunc = Callable[[Any], Any]
type PrintFunc = Callable[..., None]
type VisualFunc = Func


class TaskStop(Exception):
    pass


class Task(NamedTuple):
    stop: Event
    func: Func
    payload: Any
    pickable: Callable
    reraise: bool
    args: Iterable[Any]
    kwargs: Mapping[str, Any]


def taskproc(task: Task) -> Result:
    if task.stop.is_set():
        return Result(
            task.stop,
            task.payload,
            exception=InterruptedError("stopped"),
        )

    result = Result(task.stop, task.payload)
    outcome: Any = None
    elapsed: float = 0.0
    prev_limit = sys.getrecursionlimit()
    sys.setrecursionlimit(2**16)
    try:
        start_time = time.thread_time()
        try:
            outcome = task.func(task.payload, *task.args, **task.kwargs)
        except TypeError:
            if not isinstance(task.payload, VisualPayload):
                raise

            # HACK add backwards compatibility
            outcome = task.func(task.payload.path, *task.args, **task.kwargs)

        elapsed = time.thread_time() - start_time
    except KeyboardInterrupt:
        task.stop.set()
        raise
    except RuntimeError:
        raise
    except (Exception, RecursionError) as e:
        result.exception = e
        if task.reraise or (
            (raises := task.payload.raises())
            and not any(isinstance(e, r) for r in raises)
        ):
            raise
    finally:
        sys.setrecursionlimit(prev_limit)
        result.runtime = elapsed
        result.outcome = task.pickable(outcome)
    result.linecount = getattr(outcome, 'linecount', 0)
    result.memory = memory_use()
    return result



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# CAVEAT: THIS LEGACY MODULE IS KEPT ONLY FOR BACKWARDS COMPATIBILITY
from __future__ import annotations

from .contexts import AST
from .objectmodel import BaseNode, ModelBuilderSemantics, Node, Node as ParseModel
from .walkers import (
    BreadthFirstWalker,
    DepthFirstWalker,
    NodeWalker,
    PreOrderWalker,  # pyrefly: ignore
)


__all__ = [
    'AST',
    'BaseNode',
    'BreadthFirstWalker',
    'DepthFirstWalker',
    'ModelBuilderSemantics',
    'Node',
    'NodeWalker',
    'ParseModel',
    'PreOrderWalker',
]



from concurrent.futures import ProcessPoolExecutor, as_completed
import multiprocessing
import os
import queue
import time

# This global variable will hold the queue instance inside each worker process
_worker_queue = None

def _init_worker(q):
    """Initializes the global queue inside the child process."""
    global _worker_queue
    _worker_queue = q

def long_task(data_chunk):
    """A long-running task that emits intermediate status reports."""
    pid = os.getpid()
    counter = 0
    
    # 3-step simulated workload
    for step in range(1, 4):
        time.sleep(0.5)  # Simulate real work
        
        # Generate a lightweight, unique ID for this status update
        counter += 1
        status_id = f"{pid}-{counter}"
        
        # Send an intermediate update back to the main process
        _worker_queue.put({
            "id": status_id,
            "status": "processing",
            "msg": f"Task working on {data_chunk}: Step {step}/3 complete"
        })
        
    return f"Result for {data_chunk}"

def main():
    # Standard queue used for worker -> main communication
    status_queue = multiprocessing.Queue()
    
    # Items to process
    work_items = ["Dataset_A", "Dataset_B", "Dataset_C"]
    
    # Pass the queue to the initializer so workers can access it globally
    with ProcessPoolExecutor(initializer=_init_worker, initargs=(status_queue,)) as executor:
        
        # Submit tasks normally
        futures = {executor.submit(long_task, item): item for item in work_items}
        
        # Process results as they complete, driving the loop via the generator
        while futures:
            # 1. Drain the status queue completely to catch intermediate updates
            while True:
                try:
                    update = status_queue.get_nowait()
                    print(f"[STATUS] [ID: {update['id']}] {update['msg']}")
                except queue.Empty:
                    break
            
            # 2. Check for any finished futures without blocking indefinitely
            done_futures = [f for f in futures if f.done()]
            
            for future in done_futures:
                result = future.result()
                item = futures[future]
                print(f"--> [FINAL] {item} finished with: {result}")
                del futures[future]
            
            # Tiny sleep to keep CPU usage low while waiting for work to complete
            time.sleep(0.1)

if __name__ == "__main__":
    main()



from concurrent.futures import Executor, as_completed
import multiprocessing
import queue
import sys
from typing import Any, Callable, Iterable, Union

# The global status queue accessible by any worker process
_status_queue = multiprocessing.Queue()

def executor_pmap(
    executorcls: type[Executor],
    stop_event: multiprocessing.Event,
    process: Callable[..., Any],
    tasks: Iterable[Any],
    max_workers: int | None = None,
) -> Iterable[Union[Any, dict]]:  # Yields Results or Message dicts
    
    if not tasks:
        return

    with executorcls(max_workers=max_workers) as ex:
        try:
            futures = [ex.submit(process, task) for task in tasks]
            active_futures = set(futures)
            
            # The generator loop drives everything
            while active_futures:
                # 1. Drain the global queue and pass the "messages in a bottle" outside
                while True:
                    try:
                        # Non-blocking pull from the global queue
                        msg = _status_queue.get_nowait()
                        yield msg
                    except queue.Empty:
                        break
                
                # 2. Check for completed tasks without blocking indefinitely
                done_futures = [f for f in active_futures if f.done()]
                for future in done_futures:
                    yield future.result()  # Pass the final result outside
                    active_futures.remove(future)
                    
        except KeyboardInterrupt:
            stop_event.set()
            print(file=sys.stderr)
            print("Wait...", file=sys.stderr)
            sys.stderr.flush()
            ex.shutdown(wait=False, cancel_futures=True)
            raise



#  Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
#  SPDX-License-Identifier: BSD-4-Clause
#

# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from functools import cached_property
from typing import Any

from ..contexts import Ctx
from ..objectmodel import nodedataclass
from ..util import indent
from .base import Box, Func, Model
from .math import ffset, kdot


@nodedataclass
class Closure(Box):
    def _parse(self, ctx: Ctx) -> Any:
        return ctx.closure(self.exp._parse)

    def _first(self, k, f) -> ffset:
        efirst = self.exp._first(k, f)
        result: ffset = {()}
        for _i in range(k):
            result = kdot(result, efirst, k)
        return {()} | result

    def _pretty(self, lean=False):
        sexp = str(self.exp._pretty(lean=lean))
        if len(sexp.splitlines()) <= 1:
            return f'{{{sexp}}}'
        else:
            return f'{{\n{indent(sexp)}\n}}'

    @cached_property
    def _nullable(self) -> bool:
        return True


@nodedataclass
class PositiveClosure(Closure):
    def _parse(self, ctx: Ctx) -> Any:
        parse: Func = self.exp._parse  # type: ignore
        return ctx.positive_closure(parse)

    def _first(self, k, f) -> ffset:
        return self.exp._first(k, f)

    def _pretty(self, lean=False):
        return super()._pretty(lean=lean) + '+'

    @cached_property
    def _nullable(self) -> bool:
        return self.exp.is_nullable()


@nodedataclass
class Join(Box):
    JOINOP = '%'

    sep: Model = Model()

    def __post_init__(self):
        super().__post_init__()
        assert not self.ast or self.sep == self.ast.sep, self.sep

    def _parse(self, ctx: Ctx) -> Any:
        return self._do_parse(ctx, self.exp._parse, self.sep._parse)

    def _do_parse(self, ctx: Ctx, exp: Func, sep: Func) -> Any:
        return ctx.join(exp, sep)

    def _pretty(self, lean=False):
        ssep = self.sep._pretty(lean=lean)
        sexp = str(self.exp._pretty(lean=lean))
        if len(sexp.splitlines()) <= 1:
            return f'{ssep}{self.JOINOP}{{{sexp}}}'
        else:
            return f'{ssep}{self.JOINOP}{{\n{sexp}\n}}'

    @cached_property
    def _nullable(self) -> bool:
        return True


class PositiveJoin(Join):
    def _first(self, k, f) -> ffset:
        return self.exp._first(k, f)

    def _do_parse(self, ctx, exp, sep):
        return ctx.positive_join(exp, sep)

    def _pretty(self, lean=False):
        return super()._pretty(lean=lean) + '+'

    @cached_property
    def _nullable(self) -> bool:
        return self.exp._nullable


class Gather(Join):
    JOINOP = '.'

    def _do_parse(self, ctx, exp, sep):
        return ctx.gather(exp, sep)


class PositiveGather(Gather):
    def _first(self, k, f) -> ffset:
        return self.exp._first(k, f)

    def _do_parse(self, ctx, exp, sep):
        return ctx.positive_gather(exp, sep)

    def _pretty(self, lean=False):
        return super()._pretty(lean=lean) + '+'

    @cached_property
    def _nullable(self) -> bool:
        return self.exp.is_nullable()


@nodedataclass
class EmptyClosure(Model):
    def __post_init__(self):
        super().__post_init__()
        self.ast = None

    def _parse(self, ctx: Ctx) -> Any:
        return ctx.empty()

    def _first(self, k, f) -> ffset:
        _ = k
        _ = f
        return {()}

    def _pretty(self, lean=False):
        _ = lean
        return '{}'

    @cached_property
    def _nullable(self) -> bool:
        return True



# Copyright © 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import sys
import time
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from tatsu.barz.broker import BarBroker

from .. import packetz
from ..barz import BarRow, Col, Multi
from ..config import ParserConfig
from ..exceptions import FailedParse
from ..parproc import (
    Result,
    VisualPayload,
    parproc,
)
from ..parproc.summary import show_result, show_summary
from ..peg import Grammar
from ..util.heart import Heart
from ..ztyle import Style
from .cfg import CLIConfig
from .fmt import format_result
from .lib import Results, load_grammar


class FileHeartRow(BarRow, Heart):
    def __init__(self, queue: packetz.PacketzQueue, name: str, total: int) -> None:
        self.queue = queue
        s = Style()
        white = s.bright_white().bold()
        green = s.green()
        dim = s.black().bold()

        super().__init__(
            cols=[f"   {white(name):<60s} ", Col.bar, " " * 12],
            fill="⎯⎯⎯",
            style=[green, green, dim],
            label=name,
            total=total,
            selfstop=False,
        )

    def start(self) -> None:
        super().start()

    def beat(self, mark: int, total: int) -> None:
        self.start()
        self.update(mark, total)
        self.queue.send(data=self)

    def dead(self) -> bool:
        return False


@dataclass(slots=True)
class GrammarPayload(VisualPayload):
    grammar: Grammar
    start: str
    heart: FileHeartRow
    idx: int

    def raises(self) -> tuple[type[Exception], ...]:
        return (RecursionError, FailedParse)


def run_cmd(cfg: CLIConfig) -> Results:
    start_time = time.time()

    if not cfg.grammar:
        raise ValueError("No grammar specified")

    grammarpath = Path(cfg.grammar)
    grammar = load_grammar(grammarpath)
    start = cfg.start or None

    return run_with_progress(start_time, grammar, start, cfg)


def parse_file_task(data: GrammarPayload, *_args, **_kwargs) -> Any:
    path = Path(data.path)
    text = data.payload
    grammar, start = data.grammar, data.start

    config = ParserConfig.new()
    heart = data.heart
    config.heart = heart

    relpath = path.absolute().relative_to(Path().absolute())
    config.source = str(relpath)

    heart.start()
    sys.setrecursionlimit(2**16)
    try:
        return grammar.parse(text, start=start, config=config)
    except RecursionError as e:
        return e
    finally:
        heart.beat(mark=len(text), total=len(text))
        heart.stop()


def run_with_progress(
    start_time: float,
    grammar: Any,
    start: str | None,
    cfg: CLIConfig,
) -> Results:
    inputs = cfg.inputs

    name = Path(cfg.grammar).name
    total = len(inputs)

    multi = Multi([], out=sys.stderr)
    broker = BarBroker(multi)

    s = Style()
    yellow = s.yellow()
    top_row = BarRow(
        label=name,
        cols=[Col.bar],
        fill="--.",
        style=[yellow, yellow, s],
        total=total,
    )
    multi.add_row(top_row)
    top_row.start()

    paths = [Path(input) for input in inputs]
    payloads = []
    filehearts: dict[str, FileHeartRow] = {}
    for idx, path in enumerate(paths):
        text = path.read_text()
        fh = FileHeartRow(broker.queue, path.name, len(text))
        multi.add_row(fh)
        payloads.append(
            GrammarPayload(
                path,
                text,
                grammar=grammar,  # ty: ignore
                start=start or "",
                heart=fh,
                idx=idx,
            )
        )
        filehearts[fh.id] = fh

    try:
        results = parproc(
            parse_file_task,
            payloads,
            top_row,
            parallel=True,
            reraise=False,
            # WARNING we do the summary in-process
            summary=False,
            verbose=False,
            max_workers=cfg.nproc,
        )

        def show_results(results: Iterable[Result]) -> Iterable[Result]:
            count = 0
            last = None
            for r in results:
                yield r
                count += 1  # noqa: SIM113
                top_row.update(pos=count, total=total)
                if last:
                    show_result(multi.print, last, usecolor=cfg.usecolor)
                last = r
            if last:
                show_result(multi.print, last, usecolor=cfg.usecolor)

        if not cfg.quiet or cfg.summary:
            broker.start()

        results = show_results(results)
        if cfg.summary or cfg.verbose or not cfg.quiet:
            results = show_summary(
                start_time,
                results,
                eprint=multi.print,
                usecolor=cfg.usecolor,
                verbose=cfg.verbose,
            )
        joined = list(results)
    finally:
        multi.stop()

    return [
        (
            str(r.payload.path),
            format_result(cfg, r.outcome),
        )
        for r in joined
        if not r.exception
    ]
